--------------------------------------------------------------------------
-- Assirc                                   Copyrigth ENS ULM 1996-1998 --
-- Pierre VINCENS                   V 00.000                 10/04/1996 --
-- Pierre VINCENS                   V 01.000                 06/03/1998 --
-- Pierre VINCENS                   V 01.001                 06/03/1999 --
--------------------------------------------------------------------------
-- Programme de recherche des zones de fortes homologies dans un genome
-- Le rsultat fourni est une liste de paires de rgions qui doivent   
-- etre valides par un programme d'alignement (ex: bestfit)           
-- Le principe est le suivant:                                         
--     Dans une premiere etape, on construit un tableau indexe sur les 
--     mots nuclotidiques de longueur n et dont chaque element est un 
--     pointeur vers une liste chaine donnant la position de tous les 
--     mots de squence identique  celle de l'index.                  
--     En fonction de la taille du mot et de la longueur de la sequence
--     a analyser, il est possible de dcomposer k en deux nombres k1  
--     et k2 avec k = k1 + k2. La construction du tableau se fera alors
--     pour tous les mots possible de longueur k2, un mot de taille k1 
--     tant fix. On rptera cette opration pour tous les mots      
--     possible de longueur n1. L'intert est de rduire l'espace      
--     mmoire ncessaire au stokage de cette structure. L'inconvnient
--     est dans la ncessit de parcourir plusieurs fois la squence   
--     complte. D'un point de vue logique, cette manire de procder  
--     est sans effet.                                                 
--     Le codage de cet tape est ralis dans les procdures:         
--        Get_Seeds (pas de prindexation)                             
--        Get_Seeds_By_Mask (avec prindexation)                       
--     Dans une seconde tape, on genere pour chaque indice du tableau 
--     tous les couples possibles et on teste leur adequation a etre   
--     une region potentielement homologue. La methode utilisee compare
--     la marche alatoire ralise  partir de la graine du couple    
--     pour chacune des deux sequences.
--------------------------------------------------------------------------
with Ada.Unchecked_Conversion;
with Ada.Text_Io; use Ada.Text_Io;
with Ada.Exceptions; use Ada.Exceptions;

with Type_specifications; use Type_Specifications;
with Type_Specifications.IO; use Type_Specifications.IO;
with Strlib; use Strlib;
with Strlib.IO; use Strlib.IO;
with Options_Util; use Options_Util;

with Cpu_Times; use Cpu_Times;
with Cpu_Times.Basic; use Cpu_Times.Basic;

with Sequence_Specifications; use Sequence_Specifications;
with Sequence_Specifications.Basic; use Sequence_Specifications.Basic;
with Sequence_Io; use Sequence_Io;
with Sequence_Alignment; use Sequence_Alignment;
with Sequence_Alignment.Basic; use Sequence_Alignment.Basic;

with Assirc_Specifications; use Assirc_Specifications;
with Assirc_Specifications.Basic; use Assirc_Specifications.Basic;
with Assirc_Specifications.Database; use Assirc_Specifications.Database;
with Assirc_Exceptions; use Assirc_Exceptions;
with Assirc_Statistics; use Assirc_Statistics;
with Assirc_Global_Variables; use Assirc_Global_Variables;
with Assirc_Options; use Assirc_Options;
with Assirc_Read; use Assirc_Read;
with Assirc_Database_Functions; use Assirc_Database_Functions; 
with Assirc_Database; use Assirc_Database;
with Assirc_Database.Alignment; use Assirc_Database.Alignment;
with Assirc_Seeds; use Assirc_Seeds;
with Assirc_Seeds.Eval; use Assirc_Seeds.Eval;

procedure Assirc is

--------------------------------------------------------------------------
-- Options d'optimisation a positionner en fonction du compilateur      --
--------------------------------------------------------------------------
  --pragma Suppress(All_Checks);
  --pragma Optimize(Time);
  
--------------------------------------------------------------------------
-- Dclaration des variables locales  la procdure principale          --
--------------------------------------------------------------------------
  Display : Output_Display := File_Only;
  -- Controle de l'affichage des paires alignes  l'cran
  
--------------------------------------------------------------------------
-- Impression des statistiques de l'avance du programme                --
--------------------------------------------------------------------------
  package Duration_IO is new Ada.Text_Io.Fixed_Io(Duration);
  use Duration_IO,Int_Io;
  
  procedure Print_Statistics;
  pragma Inline(Print_Statistics);
  procedure Print_Statistics is
    Constituted, Tested   : Natural;
    Registered, Validated : Natural;
    Times : Cpu_Duration := Elapsed_Times;
  begin
    Statistics.Get_Statistics(Constituted,Tested,Registered,Validated);
    Put(Times.User_Time,6,1); Put("u ");
    Put(Times.System_Time,6,1); Put("s ");
    Put(Times.Real_Time,6,1); Put("r ");
    Put(Constituted,10); Put("C ");
    Put(Tested,10); Put("T ");
    Put(Registered,7); Put("R ");
    Put(Validated,7); Put("V");
    New_Line;
  end Print_Statistics;

--------------------------------------------------------------------------
-- Recherche des rgions similaires                                     --
--------------------------------------------------------------------------
  procedure Build_Similar_Regions
    (Sequences    : in Assirc_Sequence_Array_Access;
     Seed_Size    : in Integer;
     Rec_Length   : in Integer;
     Distance     : in Integer;
     Max_Com_Dist : in Integer;
     Tolerance    : in Integer;
     Skip         : in Integer;
     Xonly        : in Boolean);
  pragma Inline(Build_Similar_Regions);
  procedure Build_Similar_Regions
    (Sequences    : in Assirc_Sequence_Array_Access;
     Seed_Size    : in Integer;
     Rec_Length   : in Integer;
     Distance     : in Integer;
     Max_Com_Dist : in Integer;
     Tolerance    : in Integer;
     Skip         : in Integer;
     Xonly        : in Boolean) is
    Seeds : Seed_Segments;
  begin
    Init_Seeds(Seed_Size,Seeds);
    for I in Sequences'Range loop
      if Sequences(I).Strand = Watson or Sequences(I).Strand = Both then
        Get_Seeds(Sequences(I),Watson,Seed_Size,Seeds);
      end if;
      if Sequences(I).Strand = Crick or Sequences(I).Strand = Both then
        Get_Seeds(Sequences(I),Crick,Seed_Size,Seeds);
      end if;
    end loop;
    if Options.Verbose then
      Put_Line("Beginning for searching the regions of similarity");
    end if;
    Eval_Seeds(Sequences,Seeds,
               Rec_Length,Distance,
               Tolerance,Max_Com_Dist,
               Skip,Xonly);
    if Options.Verbose then
      Put_Line("End for searching the regions of similarity");
      Print_Statistics;
    end if;
  end Build_Similar_Regions;
 
  procedure Build_Similar_Regions_By_Mask
    (Sequences    : in Assirc_Sequence_Array_Access;
     Seed_Size    : in Integer;
     Mask_Size    : in Integer;
     Rec_Length   : in Integer;
     Distance     : in Integer;
     Max_Com_Dist : in Integer;
     Tolerance    : in Integer;
     Skip         : in Integer;
     Xonly        : in Boolean);
  pragma Inline(Build_Similar_Regions_By_Mask);
  procedure Build_Similar_Regions_By_Mask
    (Sequences    : in Assirc_Sequence_Array_Access;
     Seed_Size    : in Integer;
     Mask_Size    : in Integer;
     Rec_Length   : in Integer;
     Distance     : in Integer;
     Max_Com_Dist : in Integer;
     Tolerance    : in Integer;
     Skip         : in Integer;
     Xonly        : in Boolean) is
    Seeds     : Seed_Segments;
    Nb_Parts  : constant Integer := (Nb_Symbol ** Mask_Size)-1;
  begin
    Init_Seeds(Seed_Size-Mask_Size,Seeds);
    for Mask in 0..Nb_Parts loop
      for I in Sequences'range loop
        if Sequences(I).Strand = Watson or Sequences(I).Strand = Both then
          Get_Seeds_By_Mask(Sequences(I),Watson,Seed_Size,
                            Mask_Size,Mask,Seeds);
        end if;
        if Sequences(I).Strand = Crick or Sequences(I).Strand = Both then
          Get_Seeds_By_Mask(Sequences(I),Crick,Seed_Size,
                            Mask_Size,Mask,Seeds);
        end if;
      end loop;
      if Options.Verbose then
	Put_Line("Beginning for searching the regions of similarity (Mask value ="
		 & Integer'Image(Mask) & ")");
      end if;
      Eval_Seeds(Sequences,Seeds,
                 Rec_Length,Distance,
                 Tolerance,Max_Com_Dist,
                 Skip,Xonly);
      Reset_Seeds(Seeds);
      if Options.Verbose then
	Print_Statistics;
      end if;
    end loop;
    if Options.Verbose then
      Put_Line("End for searching the regions of similarity");
    end if;
  end Build_Similar_Regions_By_Mask;

begin
  -- Analyse des options
  Get_Options(Options);
  if Options.Summary then
    -- Affichage des paires a l'cran et dans un fichier
    Display := Assirc_Database.Alignment.File_And_Screen; 
  end if;
  -- Lecture et preparation des sequences
  if Length(Options.Name) = 0 then
    -- Les noms des sequences sont attendus sur la ligne de commande
    -- Initialiser la table des squences
    Sequences := new 
      Assirc_Sequence_Array(Accession_Type(1)..Accession_Type(Nb_Sequences));
    -- Lire la premire sequence
    Read_One_Sequence(To_String(Options.Name1), Options.Identifier1, 1,
                      Options.Format1, Options.Strand1, 
                      Options.Start1, Options.Stop1,
                      Sequences(1),
                      Options.Opposite1,Options.Complementary1,
		      Options.Verbose);    
    if Nb_Sequences = 1 then
      if Length(Options.Output) = 0 then
        Options.Output := Sequences(1).Identifier;
      end if;
    else
      -- Lire la deuxime sequence
      Read_One_Sequence(To_String(Options.Name2), Options.Identifier2, 2,
                        Options.Format2, Options.Strand2,
			Options.Start2, Options.Stop2,
                        Sequences(2),
                        Options.Opposite2,Options.Complementary2,
			Options.Verbose);
      if Length(Options.Output) = 0 then
        Options.Output := Sequences(1).Identifier & "_" & 
	  Sequences(2).Identifier;
      end if;
    end if;
  else
    -- Les noms des squences sont dans un fichier
    Sequences := Read_List_Of_Sequences(To_String(Options.Name),
					Options.Verbose);
    if Length(Options.Output) = 0 then
      Options.Output := To_Varstring(Root_Name(Options.Name));
    end if;
  end if;
  
  -- Positionner les parametres de gestion de la base de donnes
  -- On le fait ici pour etre sur que la base de donnes soit
  -- fonctionnelle avant l'activation des recherches.
  Assirc_Database.Set_Parameters(Options.Excluded_Marge,
                                 Options.Intersection_Marge);

  -- Positionner la table des shifts pour la marche alatoire
  Shift := Define_Shift(Options.Shift,
			To_String(Options.Shift_Name));
  Set_Seed_Parameters(Shift.all);

  -- Rechercher les squences similaires en fonction de la graine
  if Options.Mask_Size = 0 then
    Build_Similar_Regions(Sequences,
                          Options.Seed_Size,
                          Options.Recorded_Length,
                          Options.Distance,
                          Options.Common_Max_Dist,
                          Options.Tolerance,
                          Options.Skip,
                          Options.Xonly);
  else
    Build_Similar_Regions_By_Mask(Sequences,
                                  Options.Seed_Size,
                                  Options.Mask_Size,
                                  Options.Recorded_Length,
                                  Options.Distance,
                                  Options.Common_Max_Dist,
                                  Options.Tolerance,
                                  Options.Skip,
                                  Options.Xonly);
  end if;

  -- Imprimer les resultats
  Assirc_Database.Print_Database(To_String(Options.Output),
				 Options.Final_Length);

  -- Valider les rsultats 
  if Options.Alignment then
    if Options.Verbose then
      Put_Line("Alignment...");
    end if;
    Perform_Alignments(To_String(Options.Output),
		       Display,
		       Options.Final_Length,
		       Percentage(Options.Min_Similarity),
		       Options.Alignment_Method,
		       Options.Mismatch_Weight, 
		       Options.Match_Weight,
		       Options.Gap_Weight,
		       Options.Gap_Length_Weight,
		       Options.With_End_Weight,
		       Options.Max_Length,
		       Options.Adjust,
		       Options.Similarity_Threshold,
		       Options.Randomization,
		       Options.Preserve);
  end if;

  -- Imprimer les statistiques de la recherche
  if Options.Verbose then
    Print_Statistics;
  end if;
exception
  when ARGUMENT_ERROR => 
    Print_Options;
  when The_Error : others =>
    Check_Exception(The_Error);
end Assirc;

