--------------------------------------------------------------------------
-- Dassirc_Manager                          Copyrigth ENS ULM 1996-1998 --
-- Pierre VINCENS                   V 00.000                 10/04/1996 --
-- Pierre VINCENS                   V 01.000                 06/03/1998 --
--------------------------------------------------------------------------
-- Programme de recherche des zones de fortes homologies dans un genome --
-- Le rsultat fourni est une liste de paires de rgions qui doivent    --
-- etre valides par un programme d'alignement (ex: bestfit)            --
-- Le principe est le suivant:                                          --
--     Dans une premiere etape, on construit un tableau indexe sur les  --
--     mots nuclotidiques de longueur k et dont chaque element est un  --
--     pointeur vers une liste chaine donnant la position de tous les  --
--     mots de squence identique  celle de l'index.                   --
--     En fonction de la taille du mot et de la longueur de la sequence --
--     a analyser, il est possible de dcomposer k en deux nombres k1   --
--     et k2 avec k = k1 + k2. La construction du tableau se fera alors --
--     pour tous les mots possible de longueur k2, un mot de taille k1  --
--     tant fix. On rptera cette opration pour tous les mots       --
--     possible de longueur k1. L'intert est de rduire l'espace       --
--     mmoire ncessaire au stokage de cette structure. L'inconvnient --
--     est dans la ncessit de parcourir plusieurs fois la squence    --
--     complte. D'un point de vue logique, cette manire de procder   --
--     est sans effet.                                                  --
--     Le codage de cet tape est ralis dans les procdures:          --
--        Genere_Seeds (pas de prindexation)                           --
--        Genere_Seeds_By_Mask (avec prindexation)                     --
--     Dans une seconde tape, on genere pour chaque indice du tableau  --
--     tous les couples possibles et on teste leur adequation a etre    --
--     une region potentielement homologue. La methode utilisee compare --
--     la marche alatoire ralise  partir de la graine du couple     --
--     pour chacune des deux sequences.
--------------------------------------------------------------------------
with Ada.Unchecked_Conversion;
with Ada.Text_Io; use  Ada.Text_Io;

with Type_specifications; use Type_Specifications;
with Type_Specifications.IO; use Type_Specifications.IO;
with Cpu_Times.Basic; use Cpu_Times.Basic;
with Cpu_Times.Counters; use Cpu_Times.Counters;
with Strlib; use Strlib;
with Strlib.IO; use Strlib.IO;
with Options_Util; use Options_Util;

with Sequence_Specifications; use Sequence_Specifications;
with Sequence_Specifications.Basic; use Sequence_Specifications.Basic;
with Sequence_Io; use Sequence_Io;

with Assirc_Specifications; use Assirc_Specifications;
with Assirc_Specifications.Basic; use Assirc_Specifications.Basic;
with Assirc_Exceptions; use Assirc_Exceptions;
with Assirc_Global_Variables; use Assirc_Global_Variables;
with Assirc_Options; use Assirc_Options;
with Assirc_Read; use Assirc_Read;

with Dassirc_Statistics; use Dassirc_Statistics;
with Dassirc_Statistics_Io; use Dassirc_Statistics_Io;

with Dassirc_Dispatcher;
with Dassirc_Database;

procedure Dassirc_Manager is

--------------------------------------------------------------------------
-- Options d'optimisation a positionner en fonction du compilateur      --
--------------------------------------------------------------------------
  --pragma Suppress(All_Checks);
  --pragma Optimize(Time);

--------------------------------------------------------------------------
-- Acces aux bibliothques                                              --
--------------------------------------------------------------------------
  use Type_Specifications.IO.Int_Io;
  use Type_Specifications.IO.Real_Io;
  use Type_Specifications.Real_Lib;


begin
  -- Enregistrement de la tache
  Cpu_Counters.Add("Manager");
  -- Analyse des options
  Get_Options(Options);
  if Length(Options.Name) = 0 then
    -- Les noms des sequences sont attendus sur la ligne de commande
    -- Initialiser la table des squences
    Sequences := new 
      Assirc_Sequence_Array(Accession_Type(1)..Accession_Type(Nb_Sequences));
    -- Lire la premire sequence
    Read_One_Sequence(To_String(Options.Name1), Options.Identifier1, 1,
                      Options.Format1, Options.Strand1, 
		      Options.Start1, Options.Stop1,
                      Sequences(1),
                      Options.Opposite1,Options.Complementary1);    
    if Nb_Sequences = 1 then
      if Length(Options.Output) = 0 then
        Options.Output := Sequences(1).Identifier;
      end if;
    else
      -- Lire la deuxime sequence
      Read_One_Sequence(To_String(Options.Name2), Options.Identifier2, 2,
                        Options.Format2, Options.Strand2,
			Options.Start2, Options.Stop2,
                        Sequences(2),
                        Options.Opposite2,Options.Complementary2);
      if Length(Options.Output) = 0 then
        Options.Output := Sequences(1).Identifier & "_" & 
	  Sequences(2).Identifier;
      end if;
    end if;
  else
    -- Les noms des squences sont dans un fichier
    Sequences := Read_List_Of_Sequences(To_String(Options.Name));
    if Length(Options.Output) = 0 then
      Options.Output := To_Varstring(Root_Name(Options.Name));
    end if;
  end if;
  -- Initialisation des parametres du shift
  Shift := Define_Shift(Options.Shift,
			To_String(Options.Shift_Name));

  -- Initialiser la base de donnes et positionner les parametres de gestion 
  -- de la base de donnes
  -- On le fait ici pour etre sur que la base de donnes soit
  -- fonctionnelle avant l'activation des recherches.
  Put_Line("Starting Dassirc_Database");
  Dassirc_Database.Start(Options.Excluded_Marge,
			 Options.Intersection_Marge);
  Put_Line("Dassirc_Database started");
  
  -- Initialiser le dispatcher  
  Put_Line("Starting Dassirc_Dispatcher");
  Dassirc_Dispatcher.Start;
  Put_Line("Dassirc_Dispatcher started");
  
  -- Initialiser le module de gestion des statistiques 
  Put_Line("Starting Dassirc_Statistics");
  Dassirc_Statistics.Start;
  Put_Line("Dassirc_Statistics started");

  -- Faire la mise  jour de la liste des process
  Cpu_Counters.Update_Childs;
  
  -- Attendre la fin des calculs de localisation
  Dassirc_Dispatcher.Wait_End_Of_Search;
  
  -- Imprimer les resultats 
  -- Cela provoque l'attente de la fin des calculs
  Dassirc_Dispatcher.Print_Database(To_String(Options.Output),
				    Options.Final_Length);

  -- Afficher le temps de calcul de cette partition
  Print_Times;

  -- Imprimer les temps d'execution tache par tache avant l'alignement
  Put_Line("********************");
  Put_Line("Searching time report:");
  Display(Dassirc_Statistics.State);
  Put_Line("********************");
  
  -- Lancer l'alignement
  Put_Line("Manager: Start_Alignment");
  Dassirc_Dispatcher.Start_Alignment;

  -- Faire la mise  jour de la liste des process
  Put_Line("Manager: Update_Childs");
  Cpu_Counters.Update_Childs;
  Put_Line("Manager: Updated_Childs");
  
  -- Realiser l'impression
  Put_Line("Manager: Print_Valid_Pairs");
  Dassirc_Dispatcher.Print_Valid_Pairs(To_String(Options.Output));
  -- Librer la place
  Put_Line("Manager: Stop_Alignment");
  Dassirc_Dispatcher.Stop_Alignment;

  -- Afficher le temps de calcul de cette partition
  Print_Times;

  -- Indiquer a la base de donnes que tout est fini
  Dassirc_Database.Stop;
  Put_Line("Dassirc_Database stopped");

  -- Attendre la deconnexion du dernier client
  Dassirc_Dispatcher.Wait_The_Last_Client;

  -- Imprimer les temps d'execution finaux tache par tache
  Put_Line("********************");
  Put_Line("Final time report:");
  Display(Dassirc_Statistics.State);
  Put_Line("********************");
  
  -- Indiquer au dispatcher que c'est fini
  Dassirc_Dispatcher.Stop;
  -- Indiquer au module de statistiques que c'est fini
  Dassirc_Statistics.Stop;
  -- Clore le compteur local
  Cpu_Counters.Remove;
  --Cpu_Counters.Print_Process_List;
  --Cpu_Counters.Print_Elapsed_Times;
  Put_Line("Dassirc_Dispatcher stopped");

exception
  when ARGUMENT_ERROR => 
    Cpu_Counters.Remove;
    Cpu_Counters.Print_Process_List;
    Print_Options;
    raise;
  when The_Error : others =>
    Cpu_Counters.Remove;
    Cpu_Counters.Print_Process_List;
    Check_Exception(The_Error);
    raise;
end Dassirc_Manager;

