--------------------------------------------------------------------------
-- Dassirc_Dispatcher                       Copyrigth ENS ULM 1996-2000 --
-- Pierre VINCENS                   V 01.000                 29/11/2000 --
--------------------------------------------------------------------------
with Cpu_Times.Counters; use Cpu_Times.Counters;

with Ada.Text_Io; use Ada.Text_Io;

pragma Elaborate_All(Cpu_Times.Counters);

package body Dassirc_Statistics is

  type Client_Statistics is record
    Connect  : Boolean := False;         -- Etat du client
    State    : Partition_State;          -- Etat de la partition
  end record;
  type Clients_Array is array(Client_Id) of Client_Statistics;
  
  type Dispatcher_Statistics is record
    State    : Partition_State;          -- Etat de la partition
  end record;

  type Database_Statistics is record
    State    : Partition_State;          -- Etat de la partition
  end record;
  
--------------------------------------------------------------------------
-- Procdures de reprises des exceptions (pb des types proteges)        --
--------------------------------------------------------------------------
  type Exception_Id is 
    (Null_Exception,Not_Connected_Exception,Invalid_View_Exception);
  procedure Reraise_Exception(Id : in Exception_Id) is
  begin
    case Id is
      when Null_Exception              => return;
      when Not_Connected_Exception     => raise NOT_CONNECTED;
      when Invalid_View_Exception      => raise INVALID_VIEW;
    end case;
  end Reraise_Exception;
   
--------------------------------------------------------------------------
-- Type protg pour mmoriser les statistiques
--------------------------------------------------------------------------
  protected Statistics is
    -- Pour gerer les statistiques d'avancees du programme
    procedure Advancement(View   : in View_Id; 
			  State  : in Partition_State;
			  Error : out Exception_Id);
    -- pour prciser o en est l'avancement d'une recherche
    procedure Connect(View : in View_Id);
    -- Indique la connexion d'un nouveau client 
    procedure Disconnect(View  : in View_Id;
			 Error : out Exception_Id);
      -- Indique qu'un client a ete deconnecte
    procedure Set_Status(View   : in View_Id; 
			 Status : in Status_Type;
			 Error  : out Exception_Id);    
    -- pour positionner le status
    function Get_Status(View  : in View_Id) return Status_Type;
    -- pour obtenir le status
    function Program_State return Partition_State_Array;
    -- pour obtenir l'tat d'avancement des clients
  private
    Dispatcher : Dispatcher_Statistics; 
    -- tat du dispatcher
    Database : Database_Statistics;
    -- Etat de la base de donnes gnrale
    Clients : Clients_Array;   
  end Statistics;
    
  protected body Statistics is
    
    procedure Advancement(View  : in View_Id; 
                          State : in Partition_State;
			  Error : out Exception_Id) is
    begin
      Error := Null_Exception;
      if View in Client_Id then
	-- Verifier si le client est bien connecte
	if not Clients(Client_Id(View)).Connect then
	  --raise NOT_CONNECTED;
	  Error := Not_Connected_Exception;
	  return;
	end if;
        -- noter l'avance de ce client
        Clients(Client_Id(View)).State := State;
      elsif View = Database_Id then
        Database.State := State;
      elsif View = Dispatcher_Id then
        Dispatcher.State := State;
      else
        raise INVALID_VIEW;
      end if;
    end Advancement;
    
    procedure Connect(View : in View_Id) is
    -- Indique la connexion d'un nouveau client
    -- Ceci est normalement fait uniquement par le dispatcher
    begin
      if View in Client_Id then
	Clients(Client_Id(View)).Connect := True; 
      end if;
    end Connect;
  
    procedure Disconnect(View  : in View_Id;
			 Error : out Exception_Id) is 
      -- Indique qu'un client a ete deconnecte
      -- Ceci est normalement fait uniquement par le dispatcher
    begin
      Error := Null_Exception;
      if View in Client_Id then
	if Clients(Client_Id(View)).Connect then
	  Clients(Client_Id(View)).Connect := False;
	else
	  --raise NOT_CONNECTED;
	  Error := Not_Connected_Exception;
	  return;
	end if;
      end if;
    end Disconnect;
    
    procedure Set_Status(View   : in View_Id; 
			 Status : in Status_Type;
			 Error  : out Exception_Id) is
    begin
      Error := Null_Exception;
      if View in Client_Id then
        Clients(Client_Id(View)).State.Status := Status;
      elsif View = Database_Id then
        Database.State.Status := Status;
      elsif View = Dispatcher_Id then
        Dispatcher.State.Status := Status;
      else
        --raise INVALID_VIEW;
	Error := Invalid_View_Exception;
	return;
      end if;
    end Set_Status;
    
    function Get_Status(View  : in View_Id) return Status_Type is
      -- Retourne la valeur du status pour une partition donnee
    begin
      if View in Client_Id then
        return Clients(Client_Id(View)).State.Status;
      elsif View = Database_Id then
        return Database.State.Status;
      elsif View = Dispatcher_Id then
        return Dispatcher.State.Status;
      else
        return Unknown;
      end if;
    end Get_Status;
    
    function Program_State return Partition_State_Array is
      State : Partition_State_Array;   
    begin
      State(Dispatcher_Id).Status := Running; 
      Cpu_Counters.Elapsed_Times(State(Dispatcher_Id).Times); 
      State(Database_Id) := Database.State;
      for I in Client_Id loop
        State(I) := Clients(I).State;
      end loop;
      return State;
    end Program_State;
  end Statistics;
  
--------------------------------------------------------------------------
-- Foctions ouvertes aux clients
--------------------------------------------------------------------------
  
  procedure Start is
  begin
    -- En fait, on ne fait rien, les variables partition_state
    -- etant initialise a zero
    null;
  end Start;

  procedure Stop is
  begin
    Cpu_Counters.Remove; 
  end Stop;
  
  procedure Connect(Id : in View_Id) is
    -- Indique la connexion d'un nouveau client
    -- Ceci est normalement fait uniquement par le dispatcher
  begin
    Put_Line("Dassirc_statistics: Connexion du client "
	     & View_Id'Image(Id));
    Statistics.Connect(Id);
  end Connect;
  
  procedure Disconnect(Id : in View_Id) is 
    -- Indique qu'un client a ete deconnecte
    -- Ceci est normalement fait uniquement par le dispatcher
    Error : Exception_Id;
  begin
    Statistics.Disconnect(Id,Error);
    Reraise_Exception(Error);
  end Disconnect;
  
  procedure Set_Status(Id     : in View_Id; 
		       Status : in Status_Type) is
  -- Pour positionner le status pour une partition donnee
  -- Ceci est normalement fait uniquement par le dispatcher
    Error : Exception_Id;
  begin
    Statistics.Set_Status(Id,Status,Error);
    Reraise_Exception(Error);
  end Set_Status;
  
  function Get_Status(Id : in View_Id) return Status_Type is
  -- Retourne la valeur du status pour une partition donnee
  begin
    return Statistics.Get_Status(Id);
  end Get_Status;
  
  procedure Advancement(Id    : in View_Id; 
                        State : in Partition_State) is
    Error : Exception_Id;
  begin
    Statistics.Advancement(Id,State,Error);
    Reraise_Exception(Error);
  end Advancement;

  function State return Partition_State_Array is
  begin
    return Statistics.Program_State;
  end State;
  
begin
  Cpu_Counters.Add("Statistics");  
end Dassirc_Statistics;  
