--------------------------------------------------------------------------
-- Rb_Tree                                  Copyrigth ENS ULM 1986-1998 --
-- Pierre VINCENS                   V 01.000                 26/07/1996 --
-- Pierre VINCENS                   V 01.001                 30/03/1998 --
--------------------------------------------------------------------------
-- Gestion d'un arbre binaire bien balanc par la mthode de coloration
-- des noeuds (Red-Black Tree)
-- Algorithme decrit dans Corwen...
--------------------------------------------------------------------------
-- v1.001
-- Ajout de get_entries permettant d'obtenir l'accs aux lments de
-- l'arbre sous la forme d'un tableau.
--------------------------------------------------------------------------
generic
  type Element_Type is private;
  with function "="(A,B : in Element_Type) return boolean;
  with function "<"(A,B : in Element_Type) return boolean;
package Rb_Tree is

  -- L'affectation d'objet des types suivant doit etre faite avec
  -- precautions. Il s'agit d'adresse vers des objets. Ces adresses
  -- sont amenees a evoluer au cours du temps. La declaration de ces
  -- objets comme limited private limiterait ces risques mais
  -- introduit plusieurs contraintes majeures.
  type Tree_Handler is private;
  type Tree_Ptr is private;
  type Tree_Entries_Array is array(Integer range <>) of Tree_Ptr;
  type Tree_Entries is access Tree_Entries_Array;

  Null_Ptr : constant Tree_Ptr;

  type Nodecolor is (Red,Black);

  procedure Insert(Tree    : in out Tree_Handler;
                   Element : in Element_Type);
  pragma Inline(Insert);
    -- Ajoute un element dans l'arbre

  function Consult(Address : in Tree_Ptr) return Element_Type;
  pragma Inline(Consult);
    -- Retourne l'element dont l'adresse a ete specifiee

  procedure Reset(Tree : in out Tree_Handler);
  pragma Inline(Reset);
    -- Dtruit tous les noeuds de l'arbre
    -- ATTENTION, cela ne libere pas la place occupee par les elements
    -- eux memes qui seraient pointes par la structure, mais seulement
    -- la place occupe par les noeuds.

  generic
    with procedure Clean(Element : in out Element_Type);
  procedure Clean_And_Reset(Tree : in out Tree_Handler);
  pragma Inline(Clean_And_Reset);
    -- Dtruit tous les noeuds de l'arbre
    -- la procedure clean est appelee sur chaque element avant l'limination
    -- du noeud le portant ce qui permet de detruire les elements si necessaire

  procedure Delete(Tree    : in out Tree_Handler;
                   Address : in out Tree_Ptr);
  pragma Inline(Delete);
    -- Supprime l'element dont l'adresse a ete specifie

  function Find(Tree    : in Tree_Handler;
                Element : in Element_Type) return Tree_Ptr;
  pragma Inline(Find);
    -- Retourne l'adresse de l'element donne en argument
  
  type Order is (Less,Equal,More);
  generic
    type Data_Feature is limited private;
    with function Feature(Element : in Element_Type;
			  Key     : in Data_Feature) return Order;
  function Find_Element(Tree      : in Tree_Handler;
                        Key       : in Data_Feature) return Tree_Ptr;
  pragma Inline(Find_Element);
    -- Recherche le premier element verifiant la condition dfinie
    -- par le critre spcifi. Un pointeur null est retourne si pas trouve.
    -- ATTENTION, la fonction Feature retourne un type qui dit
    -- si l'lment a t trouv (equal), si l'on doit se dplacer
    -- dans la branche gauche (less) ou droite (more).

  generic
    type Data_Feature is private;
    with function Feature(Element : in Element_Type;
                          Data    : in Data_Feature) return Boolean;
  function Find_Between(Tree      : in Tree_Handler;
                        Condition : in Data_Feature;
                        Low,Hight : in Element_Type) return Tree_Ptr;
  pragma Inline(Find_Between);
    -- Recherche le premier element verifiant la condition compris
    -- entre les deux bornes low and hight
    -- L'arbre est parcouru en visitant le fils gauche, puis en appliquant
    -- command sur le noeud, puis en visitant le fils droit

  generic
    with procedure Command(Element : in out Element_Type);
  procedure Apply(Tree : in Tree_Handler);
  pragma Inline(Apply);
    -- Applique la procedure commande specifiee a l'ensemble des elements
    -- de la liste.
    -- ATTENTION, la modification de la valeur d'un element ne doit pas
    -- affecter son ordonnancement sous peine de rendre inopperante
    -- les fonctions de recherche.
    -- L'arbre est parcouru en visitant le fils gauche, puis en appliquant
    -- command sur le noeud, puis en visitant le fils droit

  generic
    type Data_Command is limited private;
    with procedure Command(Element : in out Element_Type;
                           Data    : in out Data_Command);
  procedure Apply_With_Data(Tree : in Tree_Handler;
                            Data : in out Data_Command);
  pragma Inline(Apply_With_Data);
    -- Applique la procedure commande specifiee a l'ensemble des elements
    -- de la liste. Le paramtre data est limit.
    -- ATTENTION, la modification de la valeur d'un element ne doit pas
    -- affecter son ordonnancement sous peine de rendre inopperante
    -- les fonctions de recherche.
    -- L'arbre est parcouru en visitant le fils gauche, puis en appliquant
    -- command sur le noeud, puis en visitant le fils droit;

  generic
    type Data_Command is limited private;
    with procedure Command(Element : in out Element_Type;
                           Data    : in out Data_Command);
  procedure Apply_Between(Tree : in Tree_Handler;
                          Data : in out Data_Command;
                          Low,Hight : in Element_Type);
  pragma Inline(Apply_Between);
    -- Applique la procedure commande specifiee aux elements de la liste
    -- compris entre les deux bornes low and hight
    -- ATTENTION, la modification de la valeur d'un element ne doit pas
    -- affecter son ordonnancement sous peine de rendre inopperante
    -- les fonctions de recherche.
    -- L'arbre est parcouru en visitant le fils gauche, puis en appliquant
    -- command sur le noeud, puis en visitant le fils droit

  generic
    type Data_Command is limited private;
    with procedure Command(Element : in out Element_Type;
                           Data    : in out Data_Command);
    with function Feature(Element : in Element_Type;
                          Data    : in Data_Command) return Boolean;
  procedure Apply_Between_Until(Tree      : in Tree_Handler;
                                Data      : in out Data_Command;
                                Low,Hight : in Element_Type);
  pragma Inline(Apply_Between_Until);
    -- Applique la procedure commande specifiee aux elements de la liste
    -- compris entre les deux bornes low and hight jusqu'a ce que la
    -- condition soit remplie une premiere fois. La procedure command
    -- est appliquee avant le test de la condition donc au moins une fois
    -- ATTENTION, la modification de la valeur d'un element ne doit pas
    -- affecter son ordonnancement sous peine de rendre inopperante
    -- les fonctions de recherche.
    -- L'arbre est parcouru en visitant le fils gauche, puis en appliquant
    -- command sur le noeud, puis en visitant le fils droit

  generic
    type Data_Command is private;
    type Data_Feature is private;
    with procedure Command(Element : in out Element_Type;
                           Data    : in out Data_Command);
    with function Feature(Element : in Element_Type;
                          Data    : in Data_Feature) return Boolean;
  procedure Apply_Between_If_To_First(Tree      : in Tree_Handler;
                                      Condition : in Data_Feature;
                                      Data      : in out Data_Command;
                                      Low,Hight : in Element_Type;
                                      Apply     : out Boolean);
  pragma Inline(Apply_Between_If_To_First);
    -- Applique la procedure commande specifiee aux elements de la liste
    -- compris entre les deux bornes low and hight jusqu'a ce que la
    -- condition soit remplie une premiere fois. La procedure command
    -- est appliquee apres le test de la condition.
    -- ATTENTION, la modification de la valeur d'un element ne doit pas
    -- affecter son ordonnancement sous peine de rendre inopperante
    -- les fonctions de recherche.
    -- L'arbre est parcouru en visitant le fils gauche, puis en appliquant
    -- command sur le noeud, puis en visitant le fils droit

  function Is_Empty(Tree : in Tree_Handler) return Boolean;
  pragma Inline(Is_Empty);
    -- tester si l'arbre est vide

  function Count(Tree : in Tree_Handler) return Natural;
  pragma Inline(Count);
    -- Retourne le nombre d'elements presents dans l'arbre.

  procedure Get_Entries(Tree : in Tree_Handler; Entries : in out Tree_Entries);
  pragma Inline(Get_Entries);
    -- construit un tableau contenant la liste des adresses des elements
    -- ATTENTION: ce tableau n'a de sens que tant qu'aucun element de 
    -- l'arbre n'a ete modifie.

  procedure Free(Entries : in out Tree_Entries);
  pragma Inline(Free);
    -- libere la place mmoire du tableau associ. Les lments de
    -- l'arbre sont conservs intacts.

private
  --type Nodecolor is (Red,Black);
  type Tree_Node;
  type Tree_Ptr is access Tree_Node;
  type Tree_Node is record
    Left,Right : Tree_Ptr   := null;
    Parent     : Tree_Ptr   := null;
    Color      : Nodecolor  := Black;
    Element    : Element_Type;
  end record;

  Null_Ptr : constant Tree_Ptr := null;
  NIL : constant Tree_Ptr := new Tree_Node;

  type Tree_Handler is record
    Root : Tree_Ptr := NIL;
    Count : Natural := 0;
  end record;

end Rb_Tree;
