--------------------------------------------------------------------------
-- Sequence_Alignment.Global                Copyrigth ENS ULM 1996-1999 --
-- Pierre VINCENS                   V 00.000                 15/02/1999 --
-- Pierre VINCENS                   V 00.001                 01/03/2004 --
--------------------------------------------------------------------------
with Ada.Unchecked_Deallocation;

with Strlib; use  Strlib;
with Fifo;
with Statistics;

with Ada.Text_Io; use Ada.Text_Io;
with Type_Specifications.IO; use Type_Specifications.IO;
use  Type_Specifications.IO.Int_IO;
use Type_Specifications.Real_Lib;

with Sequence_Specifications; use Sequence_Specifications;
with Sequence_Specifications.Basic; use Sequence_Specifications.Basic;
with Sequence_Specifications.Utils; use Sequence_Specifications.Utils;

with Sequence_Alignment.Utils; use Sequence_Alignment.Utils;
with Sequence_Alignment.Utils.IO; use Sequence_Alignment.Utils.IO;

package body Sequence_Alignment.Global is

  INTERNAL_ERROR : exception;

  use Int_Io,Byte_Io;
  use Positions_Fifo;
  use Optimal_Path_Fifo;

--------------------------------------------------------------------------
-- Fonctions de service
--------------------------------------------------------------------------
  function Max(A,B,C : in Match_Counter) return Match_Counter is
    -- Retourne la valeur maximale des 4 lments A, B, C et D
  begin
    return Match_Counter'Max(Match_Counter'Max(A,B),C);
  end Max;
  pragma Inline(Max);


--------------------------------------------------------------------------
-- Fonction d'alignement Global de squences gnriques
--------------------------------------------------------------------------
  generic
    type Item is (<>);
    type Sequence_Array is array(Natural range <>) of Item;
    type Match_Matrix is array(Item,Item) of Match_Counter;
    with function To_Character(Value : in Item) return Character;
  procedure Gen_Global_Alignment(Sequence1            : in Sequence_Array;
				 Sequence2            : in Sequence_Array;
				 Alignments           : in out Alignments_Access;
				 Sigma                : in Match_Matrix;
				 Match_Weight         : in Integer;
				 Mismatch_Weight      : in Integer;
				 Gap_Weight           : in Integer;
				 Gap_Length_Weight    : in Integer;
				 Progression          : in Progression_Method;
				 Building             : in Building_Method;
				 Optimize             : in Optimize_Method;
				 With_End_Weight      : in Boolean;
				 Similarity_Threshold : in Integer);

  procedure Gen_Global_Alignment(Sequence1            : in Sequence_Array;
				 Sequence2            : in Sequence_Array;
				 Alignments           : in out Alignments_Access;
				 Sigma                : in Match_Matrix;
				 Match_Weight         : in Integer;
				 Mismatch_Weight      : in Integer;
				 Gap_Weight           : in Integer;
				 Gap_Length_Weight    : in Integer;
				 Progression          : in Progression_Method;
				 Building             : in Building_Method;
				 Optimize             : in Optimize_Method;
				 With_End_Weight      : in Boolean;
				 Similarity_Threshold : in Integer) is
    New_Gap_Weight  : constant Integer := Gap_Weight + Gap_Length_Weight;
    Gap_Cost        : constant Match_Counter := Match_Counter(Gap_Weight);
    Gap_Insert_Cost : constant Match_Counter := 
      Match_Counter(Gap_Weight + Gap_Length_Weight);
    Insert_Cost     : constant Match_Counter := 
      Match_Counter(Gap_Length_Weight);

    type Match_Score is array(False..True,
                              Sequence2'First..Sequence2'Last+1)
      of Match_Counter;
    Score : Match_Score;
    Previous : Boolean := False;
    Current : Boolean := True;
    -- La table contient une colonne supplmentaire initialise  zro 
    -- pour faciliter la construction de la matrice de scores
    -- La table ne contient que 2 lignes qui servent alternativement pour
    -- construire les nouveaux scores (indice current) et mmoriser la ligne
    -- des scores prcdents (indice previous)
    -- La construction du tableau se fait de bas en haut et de droite  gauche
    -- afin de gnrer la squence aligne dans le bon sens

    type Match_Direction_Array is array(Integer range <>, Integer range <>)
      of Possible_Direction;
    pragma Pack(Match_Direction_Array);
    type Match_Direction_Access is access Match_Direction_Array;
    procedure Free is new Unchecked_Deallocation(Match_Direction_Array,
						 Match_Direction_Access);
    Direction : Match_Direction_Access;
    -- Le tableau direction contient en chaque position les directions 
    -- acceptables pour reconstruire la squence. Le codage sous la
    -- forme d'un boolen permet de grer les ventuels chemins multiples
    -- en chaque position.
    type Gap_Array is array(Integer range <>,Integer range <>) of 
      Progression_Counter;
    type Gap_Access is access Gap_Array;
    pragma Pack(Gap_Array);
    procedure Free is new Unchecked_Deallocation(Gap_Array,Gap_Access);
    Gap_1,Gap_2 : Gap_Access;
    -- Les tableaux Gap_1 et Gap_2 donnent la longueur du saut  effectuer 
    -- s'il y a insertion possible respectivement sur la premire et la la 
    -- seconde squence. ATTENTION: afin de limiter la consommation
    -- d'espace mmoire, la valeur du saut est limit  255
    -- ATTENTION: afin de limiter la consommation d'espace mmoire, la valeur 
    -- du saut pouvant etre enregistr simplement est limit  255. En
    -- consquence, on utilise 2 astuces pour pouvoir mmoriser des sauts
    -- plus longs:
    --   a/ soit il n'y a pas de saut dans l'autre direction
    --      la valeur du saut est alors enregistree
    --      comme gap_oppos = saut / 256 et
    --            gap_normal = saut mod 256
    --   b/ soit il y a aussi un saut dans l'autre direction
    --      on recherche alors dans les 255 Gap_normaux prcdents
    --      une position conduisant  la meme position de saut
    -- La consultation de la valeur des gap devra dons se faire
    -- via les deux fonctions get_gap_1 et get_gap_2

    Current_Progression : Progression_Method := Progression;
    Progress : Progression_Array;
    -- Mmorise l'ordre du backtracking 
    
    function Get_Gap_1(I,J : Integer) return Integer is
    begin
      if Direction(I,J)(Insert_2) then
	-- Il y a un gap dans l'autre direction, l'enregistrement
	-- est donc limit  la valeur de gap_1
	return Integer(Gap_1(I,J));
      else
	-- La valeur a t obtenue par coenregistrement
	return Integer(Gap_2(I,J)) * 256 + Integer(Gap_1(I,J));
      end if;
    end Get_Gap_1;
    pragma Inline(Get_Gap_1);
	
    function Get_Gap_2(I,J : Integer) return Integer is
    begin
      if Direction(I,J)(Insert_1) then
	-- Il y a un gap dans l'autre direction, l'enregistrement
	-- est donc limit  la valeur de gap_2
	return Integer(Gap_2(I,J));
      else
	-- La valeur a t obtenue par coenregistrement
	return Integer(Gap_1(I,J)) * 256 + Integer(Gap_2(I,J));
      end if;
    end Get_Gap_2;

    pragma Inline(Get_Gap_2);	
    Maximal_Score : Maximal_Score_Type;
    -- La structure qui maintient les positions des scores maxima trouvs.

    Optimal_Path : Optimal_Path_Type;
      -- La pile pour mmoriser tous les chemins optimaux trouvs

  begin
    -- Initialisation de l'espace mmoire ncessaire
    Direction := new Match_Direction_Array(Sequence1'range,Sequence2'Range);
    Gap_1 := new Gap_Array(Sequence1'range,Sequence2'Range);
    Gap_2 := new Gap_Array(Sequence1'range,Sequence2'Range);

     -- Dfinition de la progression pour la remonte
    if Building = Full then
      -- dans ce mode, on gnre les 6 modles de backtracking
      -- possible sans tenir compte de l'option fournie
      Current_Progression := Progression_Method'First;
    else
      -- dans les autres cas, on ne teste que le bakctracking specifi
      Set_Progression(Progression,Progress);
    end if;
	
    -- initialisation de la dernire ligne et de la dernire colonne
    if With_End_Weight then
      Score(Previous,Score'Last(2)) := Gap_Cost;
      for J in reverse Score'First(2)..Score'Last(2)-1 loop
	Score(Previous,J) := Score(Previous,J+1) + Insert_Cost;
      end loop;
      Score(Current,Score'Last(2)) := 
	Score(Previous,Score'Last(2)) + Insert_Cost;
      Score(Previous,Score'Last(2)) := 0;
    else
      for J in reverse Score'Range(2) loop
	Score(Previous,J) := 0;
      end loop;
      Score(Current,Score'Last(2)) := 0;
    end if;
    
    -- Construction de la matrice des direction
    Build_Direction: declare
      type Line_Score is array(Score'Range(2)) of Match_Counter;
      type Progression_Array is array(Score'Range(2)) of Integer;
      F : Line_Score;
      -- Mmorisation d'une insertion sur la squence 1
      E : Match_Counter; 
      -- Mmorisation d'une insertion sur la squence 2
      Gap_E : Integer := 0; -- Mmoriser la longueur du saut
      Gap_F : Progression_Array := (others => 0);
      Max_Gap_E,Max_Gap_F : Integer := 0;
      Nb_Stop,Nb_Align,Nb_Insert_1,Nb_Insert_2 : Integer := 0;
    begin
      -- initialiser F
      -- Ce tableau permet de mmoriser le cout associ  une insertion
      -- de plusieurs nuclotides suivant la squence 1. On suppose en effet
      -- dans le modle d'alignement que l'ajout d'une nouvelle insertion
      -- est diffrent de celui d'allonger une insertion existante.
      F(F'Last) := Gap_Cost;
      for J in reverse F'First..F'Last-1 loop
	F(J) := F(J+1) + Insert_Cost;
      end loop;
      -- Calculer la matrice des directions 
      -- La construction se fait en partant de la fin des squences
      for I in reverse Sequence1'Range loop
	-- initialiser E pendant de F pour la squence 2
	E := Gap_Cost + Match_Counter(Sequence1'Last - I + 1) * Insert_Cost;
	Gap_E := 0;
        for J in reverse Sequence2'Range loop
	  -- Dterminer pour cette nouvelle case la valeur de E
	  Progression_E: declare
	    If_Add : Match_Counter := E + Insert_Cost;
	    If_New : Match_Counter := Score(Current,J+1) + Gap_Insert_Cost;
	  begin
	    if If_Add >= If_New then
	      E  := If_Add;
	      Gap_E := Gap_E + 1;
	    else
	      E  := If_New;
	      Gap_E := 1;
	    end if;
	  end Progression_E;
	  -- Dterminer pour cette nouvelle case la valeur de F
	  Progression_F: declare
	    If_Add : Match_Counter := F(J) + Insert_Cost;
	    If_New : Match_Counter := Score(Previous,J) + Gap_Insert_Cost;
	  begin
	    if If_Add >= If_New then
	      F(J)  := If_Add;
	      Gap_F(J) := Gap_F(J) + 1;
	    else
	      F(J)  := If_New;
	      Gap_F(J) := 1;
	    end if;
	  end Progression_F;
	  -- Dterminer le score 
	  Eval_Score: declare
	    If_Align : Match_Counter := 
	      Score(Previous,J+1) + Sigma(Sequence1(I),Sequence2(J));
	    S : Match_Counter renames Score(Current,J);
	  begin
	    -- on initialise tous les choix  faux
	    Direction(I,J) := (others => False);
	    -- calculer le nouveau score
	    S := Max(If_Align,E,F(J));
	    if S = If_Align then
	      Direction(I,J)(Align) := True;
	      Nb_Align := Nb_Align + 1;
	    end if;
	    if S = E then
	      Direction(I,J)(Insert_1) := True;
	      if Gap_E > 255 then
		-- On depasse la capacit du compteur pour enregistrer
		-- le saut. 
		if S = F(J) then
		  -- Il y a aussi un saut  prendre en compte
		  -- dans l'autre direction
		  Gap_1(I,J) := 0;
		  -- On regarde la plus haute valeur possible de saut
		  -- qui permette une decomposition du saut
		  declare
		    Right : constant Possible_Direction := 
		      (Stop => False, 
		       Align => False, 
		       Insert_1 => True, 
		       Insert_2 => False);
		    Dest : constant Integer := J + Gap_E;
		  begin
		    for K in reverse J+1..Integer'Min(J+255,Sequence2'Last) 
		    loop
		      if Direction(I,K) = Right and then 
			K + Get_Gap_1(I,K) = Dest
		      then
			-- Il n'y a qu'un insert_1 autoris en ce point
			-- et ce saut conduit l o il faut
			Gap_1(I,J) := Progression_Counter(K-J);
			exit;
		      end if;
		    end loop;
		    if Gap_1(I,J) = 0 then
		      -- On fixe arbitrairement le saut  255
		      -- Atteindre cette condition est quasiment impossible
		      Gap_1(I,J) := 255;
		    end if;
		  end;
		else
		  -- Il n'y a pas de saut dans l'autre direction
		  Gap_2(I,J) := Progression_Counter(Gap_E / 256);
		  Gap_1(I,J) := Progression_Counter(Gap_E mod 256);
		end if;
	      else
		-- Le saut est enregistre dans gap_1
	  	Gap_1(I,J) := Progression_Counter(Gap_E);
		-- et si pas de gap dans l'autre direction, on 
		-- enregistre une valeur nulle
	  	if S /= F(J) then
		  Gap_2(I,J) := 0;
		end if;
	      end if;
	      Nb_Insert_1 := Nb_Insert_1 + 1;
	      Max_Gap_E := Integer'Max(Max_Gap_E,Gap_E);
	    end if;
	    if S = F(J) then
	      Direction(I,J)(Insert_2) := True;
	      if Gap_F(J) > 255 then
		-- On depasse la capacit du compteur pour enregistrer
		-- le saut. 
		if S = E then
		  -- Il y a aussi un saut  prendre en compte
		  -- dans l'autre direction
		  Gap_2(I,J) := 0;
		  -- On regarde la plus haute valeur possible de saut
		  -- qui permette une decomposition du saut
		  declare
		    Up : constant Possible_Direction := (Stop => False, 
							 Align => False, 
							 Insert_1 => False, 
							 Insert_2 => True);
		    Dest : constant Integer := I + Gap_F(J);
		  begin
		    for K in reverse I+1..Integer'Min(I+255,Sequence1'Last) 
		    loop
		      if Direction(K,J) = Up and then 
			K + Get_Gap_2(K,J) = Dest 
		      then
			-- Il n'y a qu'un insert_2 autoris en ce point
			-- et ce saut conduit l o il faut
			Gap_2(I,J) := Progression_Counter(K-I);
			exit;
		      end if;
		    end loop;
		    if Gap_2(I,J) = 0 then
		      -- On fixe arbitrairement le saut  255
		      -- Condition quasiment impossible a atteindre
		      -- et si elle est elle correspond a das cas de
		      -- path sans interet
		      Gap_2(I,J) := 255;
		    end if;
		  end;
		else
		  -- Il n'y a pas de saut dans l'autre direction
		  Gap_1(I,J) := Progression_Counter(Gap_F(J) / 256);
		  Gap_2(I,J) := Progression_Counter(Gap_F(J) mod 256);
		end if;
	      else
		-- Le saut est enregistre dans gap_2
	  	Gap_2(I,J) := Progression_Counter(Gap_F(J));
		-- et si pas de gap dans l'autre direction, on 
		-- enregistre une valeur nulle
	  	if S /= E then
		  Gap_1(I,J) := 0;
		end if;
	      end if;
	      Nb_Insert_2 := Nb_Insert_2 + 1;
	      Max_Gap_F := Integer'Max(Max_Gap_F,Gap_F(J));
	    end if;
	  end Eval_Score;
        end loop;
	-- Dterminer si le score en premire position de la ligne est maximal
	declare
	  S : Match_Counter renames Score(Current,Score'First(2));
	begin
	  if S > Maximal_Score.Value then
	    -- Le score est plus grand que la dernire valeur
	    -- mmorise. On ignore les prcdents
	    Reset(Maximal_Score.Stack);
	    -- On enregistre la position et le nouveau seuil.
	    Maximal_Score.Value := S;
	    Push(Maximal_Score.Stack, (I,Sequence2'First));
	  elsif S = Maximal_Score.Value then
	    -- Le score est gal  la plus grande valeur mmorise
	    -- On ajoute  la pile la nouvelle position
	    Push(Maximal_Score.Stack, (I,Sequence2'First));
	  end if;
	end;
	-- incrmenter les compteurs pour passer  la ligne suivante
	Previous := not Previous;
	Current := not Current;
	-- Initialiser le score du dernier lment de la nouvelle ligne
	-- courante
	if With_End_Weight then
	  Score(Current,Score'Last(2)) := 
	    Score(Previous,Score'Last(2)) + Insert_Cost;
	else
	  Score(Current,Score'Last(2)) := 0;
	end if;
      end loop;
      -- Dterminer le point de dpart du chemin
      if With_End_Weight then
	-- On dmare obligatoirement du premier point des deux squences
	Reset(Maximal_Score.Stack);
	Maximal_Score.Value := Score(Previous,Sequence2'First);
	Push(Maximal_Score.Stack, (Sequence1'First,Sequence2'First));
      else
	-- Dterminer si le score en dernire ligne est maximal  
	for I in Score'First(2)+1..Score'Last(2) loop
	  declare
	    S : Match_Counter renames Score(Previous,I);
	    -- Attention, ligne previous car le swap est fait en fin de boucle
	  begin
	    if S > Maximal_Score.Value then
	      -- Le score est plus grand que la dernire valeur
	      -- mmorise. On ignore les prcdents
	      Reset(Maximal_Score.Stack);
	      -- On enregistre la position et le nouveau seuil.
	      Maximal_Score.Value := S;
	      Push(Maximal_Score.Stack, (Sequence1'First,I));
	    elsif S = Maximal_Score.Value then
	      -- Le score est gal  la plus grande valeur mmorise
	      -- On ajoute  la pile la nouvelle position
	      Push(Maximal_Score.Stack, (Sequence1'First,I));
	    end if;
	  end;
	end loop;
      end if;
    end Build_Direction;

    -- Rechercher le chemin optimum
    Optimal_Path_Searching: declare
      Path : Path_access; 
      Position : Position_Type;
      K   : Integer; -- le point de progression dans le chemin courant
      X,Y : Integer;
    begin
      -- Rserver l'espace pour mmoriser le chemin en construction.
      -- Il peut avoir au plus une longueur gale  la somme des longueurs 
      -- des deux squences (cas o il n'y a aucun match) plus 2 pour tenir 
      -- compte du traitement de la dernire position
      Path := new Path_Array(1..Sequence1'Length+Sequence2'Length+2); 
      -- Initialiser la variable maintenant le meilleur score
      -- en fonction du critre d'optimisation
      case Optimize is
	when Short => Optimal_Path.Best := Integer'Last;
	when Large => Optimal_Path.Best := Integer'First;
	when others => null;
      end case;
      -- Pour chaque dbut potentiel du chemin qui peut-etre
      -- quelquonque contrairement  l'alignement local
      All_Path_Searching: loop
	exit All_Path_Searching when Is_Empty(Maximal_Score.Stack);
	if Building = Full then 
	  if Current_Progression < Progression_Method'Last then
	    -- On consulte simplement le premier lment de la pile	  
	    Top(Maximal_Score.Stack,Position);
	  else
	    -- On dpile le premier lment de la pile, car c'est la
	    -- dernire mthode de backtracking  valuer
	    Pull(Maximal_Score.Stack,Position);
	  end if;
	  -- On dtermine l'ordre des backtracks
	  Set_Progression(Current_Progression,Progress);
	else
	  -- On dpile le premier lment de la pile
	  Pull(Maximal_Score.Stack,Position);
	end if;
	-- On initialise x et y sur le point de dpart
	X := Position.X;
	Y := Position.Y;
	-- Dans tous les cas, on considre que du premier point
	-- enregistr, il n'y a qu'un choix possible et que de ce
	-- point, il part un alignement (match ou mismatch)
	K := 1;
	-- On analyse et mmorise le premier dplacement ncessaire
	-- Si alignement, on est conduit directement au second point
	-- Si insertion, on considre celle-ci depuis l'origine
	if X = Sequence1'First and Y = Sequence2'First then
	  -- Alignement
	  Path(K).Forward := Align;
	  -- On enregistre la position de ce point initial
	  Path(K).X := X;
	  Path(K).Y := Y;
	  Path(K).Length := 1;
	  -- On calcule le dplacement pour aller au point suivant
	  X := X + 1;
	  Y := Y + 1;
	elsif X = Sequence1'First and Y > Sequence2'First then
	  -- Insertion sur la premire squence
	  Path(K).Forward := Insert_1;
	  -- On enregistre le point d'origine de la squence
	  Path(K).X := Sequence1'First;
	  Path(K).Y := Sequence2'First;
	  -- On calcule le dplacement ncessaire pour atteindre le point
	  -- de dpart depuis le dbut de la squence
	  declare  
	    Dy : constant Integer := Y - Sequence2'First;
	  begin
	    Path(K).Length := Dy;
	  end;
	  -- Le point suivant enregistr sera ce point
	elsif X > Sequence1'First and Y = Sequence2'First then
	  -- Insertion sur la premire squence
	  Path(K).Forward := Insert_2;
	  -- On enregistre le point d'origine de la squence
	  Path(K).X := Sequence1'First;
	  Path(K).Y := Sequence2'First;
	  -- On calcule le dplacement ncessaire pour atteindre le point
	  -- de dpart depuis le dbut de la squence
	  declare  
	    Dx : constant Integer := X - Sequence1'First;
	  begin
	    Path(K).Length := Dx;
	  end;
	  -- Le point suivant enregistr sera ce point
	else
	  -- Aucun choix retenu, ce qui est anormal
	  Put_Line("Error IE001");
	  raise Internal_Error;
	end if;
	-- Construction du chemin
	Path_Searching:  loop
	  -- Poursuivre le chemin en ce point
	  if X > Sequence1'Last or else Y > Sequence2'Last then
	    -- on a atteint le point terminal
	    if X <= Sequence1'Last then
	      -- On complete sur la premire squence pour atteindre 
	      -- le point terminal
	      declare
		DX : constant Integer := Sequence1'Last - X + 1;
	      begin
		K := K + 1;
		Path(K).X := X;
		Path(K).Y := Y;
		Path(K).Forward := Insert_2;
		Path(K).Length := Path(K-1).Length + DX;
		X := X + DX;
	      end;
	    elsif Y <= Sequence2'Last then
	      -- On complete sur la seconde squence pour atteindre 
	      -- le point terminal
	      declare
		DY : constant Integer := Sequence2'Last - Y + 1;
	      begin
		K := K + 1;
		Path(K).X := X;
		Path(K).Y := Y;
		Path(K).Forward := Insert_1;
		Path(K).Length := Path(K-1).Length + DY;
		Y := Y + DY;
	      end;
	    end if;
	    -- On note le point terminal du chemin comme tant 
	    -- l'extrmit des deux squences. C'est ncessaire 
	    -- pour clore la squence si on ne termine pas
	    -- sur un match
	    K := K + 1;
	    Path(K).X := X;
	    Path(K).Y := Y;
	    Path(K).Length := Path(K-1).Length;
	    Path(K).Forward := Stop;
	    -- On a atteind un bord du tableau
	    -- On mmorise le nouveau chemin s'il est meilleur
	    -- ou gal par rapport au critre d'optimisation
	    -- secondaire dfini
	    case Best_Path(Path(K),Optimal_Path.Best,Optimize) is
	      when Good =>
		-- On oublie tous les autres chemins
		Purge(Optimal_Path.Stack);
		-- et on mmorise celui-l
		declare
		  Current : Path_Record :=
		    (Path => new Path_Array(1..K), 
		     Progression => Current_Progression);
		begin
		  Current.Path.all := Path(1..K);
		  Push(Optimal_Path.Stack,Current);
		  case Optimize is
		    when Short | Large => Optimal_Path.Best := Path(K).Length;
		    when others => null;
		  end case;
		end;
	      when Equal =>
		-- C'est un chemin de meme valeur d'optimisation que celui 
		-- dj mmoris
		declare
		  Current : Path_Record :=
		    (Path => new Path_Array(1..K), 
		     Progression => Current_Progression);
		begin
		  Current.Path.all := Path(1..K);
		  if not Is_In_Path_Stack(Optimal_Path.Stack,Current) then
		    -- on ajoute ce path  la liste
		    Push(Optimal_Path.Stack,Current);
		  end if;
		end;
	      when Bad =>
		-- On ne fait rien
		null;
	    end case;
	    -- En fonction de l'optimisation voulue
	    case Building is
	      when One => 
		-- On ne recherche que le premier chemin
		exit All_Path_Searching;
	      when All_Firsts => 
		-- On ne recherche qu'un chemin par score maximal
		exit Path_Searching;
	      when Full =>
		if Current_Progression = Progression_Method'Last then
		  -- Toutes les cas possibles ont t examins
		  Current_Progression := Progression_Method'First;
		else
		  Current_Progression := 
		    Progression_Method'Succ(Current_Progression);
		end if;
		exit Path_Searching;
	    end case;
	  end if;
	  -- Dans les autres cas, on doit allonger le chemin
	  K := K + 1;
	  -- et enregistrer la nouvelle position
	  Path(K).X := X;
	  Path(K).Y := Y;
	  -- Et passer en revue les diffrents choix dans l'ordre choisi
	  -- ATTENTION, le calcul des nouvelles positions de X et de Y
	  -- doit etre fait aprs les calculs impliquant Direction(X,Y),
	  -- les valeurs de ce tableau donnant des indications pour se
	  -- rendre  l'lment suivant.
	  Path(K).Forward := Stop; -- Aucun choix retenu
	  for I in Progress'Range loop
	    if Direction(X,Y)(Progress(I)) then
	      Path(K).Forward := Progress(I);
	      exit;
	    end if;
	  end loop;
	  -- On calcule l'volution de la progression pour le choix 
	  -- qui a t retenu
	  case Path(K).Forward is
	    when Align =>
	      Path(K).Length := Path(K-1).Length + 1;
	      X := X + 1;
	      Y := Y + 1;
	    when Insert_1 =>
	      Path(K).Length := Path(K-1).Length + Get_Gap_1(X,Y);
	      Y := Y + Get_Gap_1(X,Y);
	    when Insert_2 =>
	      Path(K).Length := Path(K-1).Length + Get_Gap_2(X,Y);
	      X := X + Get_Gap_2(X,Y);
	    when others =>
	      -- Aucun choix retenu, ce qui est anormal
	      Put_Line("Error IE003");
	      raise Internal_Error;
	  end case;
	end loop Path_Searching;
      end loop All_Path_Searching;
      -- Librer la place occupe par le path
      Free(Path);
    end Optimal_Path_Searching;

    Registration: declare
      Indice : Integer := 0;
      Effective_Length : Integer; 
      -- Pour calculer la longueur sans tenir compte des gaps aux extrmits
      Ratio_Length : Integer; 
      -- Pour calculer la longueur du segment le plus court toujours sans 
      -- tenir compte des gaps aux extrmits
      Count_For_Similarity : Integer := 0;
      Local_Similarity_Threshold : Real := 0.0;
      Local_Similarity_Count : Integer := 0;
      -- Pour calculer la similitude
    begin
      -- dterminer le seuil de similitude comme la moyenne des
      -- valeurs positive de la matrice hors diagonale (= identit)
      if Similarity_Threshold >= 0 then
	Local_Similarity_Threshold := Real(Similarity_Threshold);
      else
	for I in Sigma'Range(1) loop
	  for J in Sigma'Range(2) loop
	    if I /= J and Sigma(I,J) > 0 then
	      Local_Similarity_Threshold := Local_Similarity_Threshold +
		Real(Sigma(I,J));
	      Local_Similarity_Count := Local_Similarity_Count + 1;
	    end if;
	  end loop;
	end loop;
	if Local_Similarity_Count > 0 then
	  Local_Similarity_Threshold := Local_Similarity_Threshold /
	    Real(Local_Similarity_Count); 
	else
	  Local_Similarity_Threshold := 0.0;
	end if;
      end if;
      -- creer l'espace ncessaire pour mmoriser les alignements
      Alignments := new 
	Alignments_Array(1..Optimal_Path_Fifo.Count(Optimal_Path.Stack));
      for I in Alignments'Range loop
	Alignments(I) := new Alignment_Model;
      end loop;
      while Is_Not_Empty(Optimal_Path.Stack) loop
	Indice := Indice + 1;
	One_Registration: declare
	  Alignment : Alignment_Model renames Alignments(Indice).all;
	  Data : Path_Record;
	  Path : Path_Access;
	  X : Integer;
	  Y : Integer;
	  Continuous_Match : Integer := 0;
	  I : Integer := 1; -- indice sur la squence construite
	begin
	  -- Rcuprer sur la pile un chemin
	  Pull(Optimal_Path.Stack,Data);
	  Path := Data.Path;
	  -- Mmoriser les parametres utiliss pour dterminer 
	  -- l'alignement
	  Alignment.Method := Global_Alignment;
	  Alignment.Progression := Data.Progression;
	  Alignment.Match_Weight := Integer(Match_Weight);
	  Alignment.Mismatch_Weight := Integer(Mismatch_Weight);
	  Alignment.Gap_Weight := Integer(Gap_Weight);
	  Alignment.Gap_Length_Weight := Integer(Gap_Length_Weight);
	  -- dterminer la longueur de l'alignement
	  Alignment.Length := Path(Path'Last).Length;
	  -- et dfinir les bornes initiales de la squence
	  Alignment.Start1 := Path(1).X;
	  Alignment.Start2 := Path(1).Y;
	  -- initialisation des sequences;
	  Free(Alignment.Seq1.Value);
	  Alignment.Seq1.Value := new String(1..Alignment.Length);
	  Free(Alignment.Seq2.Value);
	  Alignment.Seq2.Value := new String(1..Alignment.Length);
	  -- initialisation des mesures effectuees sur les sequences
	  Alignment.Nb_Match := 0;
	  Alignment.Nb_Mismatch := 0;
	  Alignment.Nb_Gaps := 0;
	  Alignment.Gaps_Length := 0;
	  Alignment.Max_Continuous_Match := 0;
	  -- enregistrement de la sequence
	  for K in Path'Range loop
	    X := Path(K).X;
	    Y := Path(K).Y;
	    -- enregistrement du point
	    case Path(K).Forward is
	      when Align =>
		if Real(Sigma(Sequence1(X),Sequence2(Y))) >= Local_Similarity_Threshold then
		  Count_For_Similarity := Count_For_Similarity + 1;
	        end if;
		Alignment.Seq1.Value(I) := To_Character(Sequence1(X));
		Alignment.Seq2.Value(I) := To_Character(Sequence2(Y));
		I := I + 1;
		if Sequence1(X) = Sequence2(Y) then
		  Alignment.Nb_Match := Alignment.Nb_Match + 1;
		  Continuous_Match := Continuous_Match + 1;
		else
		  Alignment.Nb_Mismatch := Alignment.Nb_Mismatch + 1;
		  if Continuous_Match > 0 then
		    Alignment.Max_Continuous_Match :=
		      Integer'Max(Continuous_Match,
				  Alignment.Max_Continuous_Match);
		    Continuous_Match := 0;
		  end if;
		end if;
	      when Insert_1 =>
		for V in Y..Path(K+1).Y-1 loop
		  Alignment.Seq1.Value(I) := '.';
		  Alignment.Seq2.Value(I) := To_Character(Sequence2(V));
		  I := I + 1;
		end loop;
		if K = 1 or else Path(K-1).Forward /= Insert_1 then
		  -- c'est un nouveau gap (1er lment ou prcdent non gap)
		  Alignment.Nb_Gaps := Alignment.Nb_Gaps + 1;
		end if;
		Alignment.Gaps_Length := 
		  Alignment.Gaps_Length+Path(K+1).Y-Y;
		if Continuous_Match > 0 then
		  Alignment.Max_Continuous_Match :=
		    Integer'Max(Continuous_Match,
				Alignment.Max_Continuous_Match);
		  Continuous_Match := 0;
		end if;
	      when Insert_2 =>
		for U in X..Path(K+1).X-1 loop
		  Alignment.Seq1.Value(I) := To_Character(Sequence1(U));
		  Alignment.Seq2.Value(I) := '.';
		  I := I + 1;
		end loop;
		if K = 1 or else Path(K-1).Forward /= Insert_2 then
		  -- c'est un nouveau gap (1er lment ou prcdent non gap)
		  Alignment.Nb_Gaps := Alignment.Nb_Gaps + 1;
		end if;
		Alignment.Gaps_Length := 
		  Alignment.Gaps_Length+Path(K+1).X-X; 
		if Continuous_Match > 0 then
		  Alignment.Max_Continuous_Match :=
		    Integer'Max(Continuous_Match,
				Alignment.Max_Continuous_Match);
		  Continuous_Match := 0;
		end if;
	      when Stop =>
		if K /= Path'Last then
		  -- Un stop en dernire position est le seul cas admis
		  Put_Line("Error IE004");
		  raise Internal_Error;
		end if;
	    end case;
	  end loop;
	  -- fin de l'alignement
	  -- ajuster le dernier continuous_match
	  Alignment.Max_Continuous_Match :=
	    Integer'Max(Continuous_Match,
			Alignment.Max_Continuous_Match);
	  -- mmoriser la position finale dans chacune des squences
	  Alignment.Stop1 := X-1;
	  Alignment.Stop2 := Y-1;
	  -- Dterminer les corrections  faire suivant si l'on tiend
	  -- compte des gaps aux extrmits ou non
	  Effective_Length := Alignment.Length;
	  if not With_End_Weight then
	    -- Si on ne tiend pas compte des gaps aux extrmits, les
	    -- supprimer
	    -- Traiter l'ventuel gap initial
	    case Path(1).Forward is
	      when Insert_1 =>
		Alignment.Nb_Gaps := Alignment.Nb_Gaps - 1;
		Alignment.Gaps_Length := 
		  Alignment.Gaps_Length-(Path(2).Y-Path(1).Y);
		Effective_Length := 
		  Effective_Length-(Path(2).Y-Path(1).Y);
	      when Insert_2 =>
		Alignment.Nb_Gaps := Alignment.Nb_Gaps - 1;
		Alignment.Gaps_Length := 
		  Alignment.Gaps_Length-(Path(2).X-Path(1).X);
		Effective_Length := 
		  Effective_Length-(Path(2).X-Path(1).X);
	      when others =>
		null;
	    end case;
	    -- Traiter l'ventuel gap final
	    declare
	      Last : constant Integer := Path'Last - 1;
	    begin
	      case Path(Last).Forward is
		when Insert_1 =>
		  Alignment.Nb_Gaps := Alignment.Nb_Gaps - 1;
		  Alignment.Gaps_Length := 
		    Alignment.Gaps_Length-(Sequence2'Last-Path(Last).Y+1);
		  Effective_Length := 
		    Effective_Length-(Sequence2'Last-Path(Last).Y+1);
		when Insert_2 =>
		  Alignment.Nb_Gaps := Alignment.Nb_Gaps - 1;
		  Alignment.Gaps_Length := 
		    Alignment.Gaps_Length-(Sequence1'Last-Path(Last).X+1);
		  Effective_Length := 
		    Effective_Length-(Sequence1'Last-Path(Last).X+1);
		when others =>
		  null;
	      end case;
	    end;
	  end if;
	  -- Dterminer le segment le plus court. On ne tiend pas compte
	  -- des lments non apparis.
	  Ratio_Length := Integer'Min(Sequence1'Length,Sequence2'Length);
	  -- calcul des valeurs de similarite and co
	  Alignment.Quality := Maximal_Score.Value;
	  Alignment.Ratio := Real(Alignment.Quality) / Real(Ratio_Length);
	  Alignment.Identity :=
	    Real(Alignment.Nb_Match*100) / 
	    Real(Effective_Length-Alignment.Gaps_Length);
	  Alignment.Similarity := 
	      Real(Count_For_Similarity*100) / 
	      Real(Effective_Length-Alignment.Gaps_Length);
	  -- Librer la place occup par ce path
	  Free(Path);
	end One_Registration;
      end loop;
    end Registration;

    -- Librer la place mmoire encore utilise
    Purge(Optimal_Path.Stack);
    Free(Direction);
    Free(Gap_1);
    Free(Gap_2);
  end Gen_Global_Alignment;

--------------------------------------------------------------------------
-- Fonction d'alignement global de squences nuclotidiques
--------------------------------------------------------------------------
  procedure Nuc_Global_Alignment is new 
    Gen_Global_Alignment(Nucleotide_Type,Nucleotidic_Array,
			 Nucleotide_Match_Matrix,
			 To_Character);

--------------------------------------------------------------------------
-- Fonction d'alignement global de squences protiques
--------------------------------------------------------------------------
  procedure Prot_Global_Alignment is new 
    Gen_Global_Alignment(Amino_Acid_Type,Amino_Acids_Array,
			 Protein_Match_Matrix,
			 To_Character);

--------------------------------------------------------------------------
-- Procedure de pretraitement  l'alignement
--------------------------------------------------------------------------
  procedure Pre_Alignment(Sequence1 : in Sequence_Data;
			  Sequence2 : in Sequence_Data;
			  From1,To1 : in Integer := Complete;
			  From2,To2 : in Integer := Complete;
			  Inf1,Sup1 : in out Integer;
			  Inf2,Sup2 : in out Integer);
  pragma Inline(Pre_Alignment);

  procedure Pre_Alignment(Sequence1 : in Sequence_Data;
			  Sequence2 : in Sequence_Data;
			  From1,To1 : in Integer := Complete;
			  From2,To2 : in Integer := Complete;
			  Inf1,Sup1 : in out Integer;
			  Inf2,Sup2 : in out Integer) is
  begin
    -- Verification de la coherence des sequences
    if Sequence1.Model /= Sequence2.Model then
      raise INCONSISTENT_MODEL;
    end if;
    -- verifier que la squence est bien sous un format nucleotide_type
    -- et initialiser les bornes
    case Sequence1.Model is
      when DNA | RNA =>
        if Sequence1.Nucleotides = null or Sequence2.Nucleotides = null then
          raise INVALID_FORMAT;
        end if;
        Inf1 := Sequence1.Nucleotides'First;
        Sup1 := Sequence1.Nucleotides'Last;
        Inf2 := Sequence2.Nucleotides'First;
        Sup2 := Sequence2.Nucleotides'Last;
      when Protein =>
        if Sequence1.Amino_Acids = null or Sequence2.Amino_Acids = null then
          raise INVALID_FORMAT;
        end if;
        Inf1 := Sequence1.Amino_Acids'First;
        Sup1 := Sequence1.Amino_Acids'Last;
        Inf2 := Sequence2.Amino_Acids'First;
        Sup2 := Sequence2.Amino_Acids'Last;
      when others =>
        raise NOT_IMPLEMENTED;
    end case;
    -- Determine les bornes du traitement et verifier leur validit
    if From1 /= Complete then Inf1 := From1; end if;
    if To1 /= Complete then Sup1 := To1; end if;
    if From2 /= Complete then Inf2 := From2; end if;
    if To2 /= Complete then Sup2 := To2; end if;
    if Inf1 >= Sup1 or Inf2 >= Sup2 then
      raise INCONSISTENT_RANGE;
    end if;
  end Pre_Alignment;

--------------------------------------------------------------------------
-- Procedure de posttraitement  l'alignement
--------------------------------------------------------------------------
  procedure Post_Alignment(Sequence1  : in Sequence_Data;
			   Sequence2  : in Sequence_Data;
			   Alignments : in out Alignments_Access;
			   Matrix     : in String  := "");
  pragma Inline(Post_Alignment);

  procedure Post_Alignment(Sequence1  : in Sequence_Data;
			   Sequence2  : in Sequence_Data;
			   Alignments : in out Alignments_Access;
			   Matrix     : in String  := "") is
  begin
    -- Complter les informations pour chaque alignement
    for I in Alignments'Range loop
      -- retenir le nom des sequences
      Copy(Sequence1.Identifier,Alignments(I).Name1);
      Copy(Sequence2.Identifier,Alignments(I).Name2);
      -- mettre a defaut le brin a Watson
      Alignments(I).Strand1 := Watson;
      Alignments(I).Strand2 := Watson;
      -- leurs types
      Alignments(I).Model := Sequence1.Model;
      -- et la matrice utilise pour faire l'appariement
      if Matrix = "" then
	Clear(Alignments(I).Matrix);
      else
	Copy(Matrix,Alignments(I).Matrix);
      end if;
    end loop;
  end Post_Alignment;

--------------------------------------------------------------------------
-- Procedure de slection
--------------------------------------------------------------------------
 
  procedure Perform_Global_Alignment
    (Sequence1            : in Sequence_Data;
     Sequence2            : in Sequence_Data;
     Alignments           : in out Alignments_Access;
     From1,To1            : in Integer := Complete;
     From2,To2            : in Integer := Complete;
     Matrix               : in String  := "";
     Match_Weight         : in Integer := 100;
     Mismatch_Weight      : in Integer := -90;
     Gap_Weight           : in Integer := -500;
     Gap_Length_Weight    : in Integer := -30;
     Progression          : in Progression_Method := DVH;
     Building             : in Building_Method := All_Firsts;
     Optimize             : in Optimize_Method := No;
     With_End_Weight      : in Boolean := False;
     Similarity_Threshold : in Integer := -1;
     Randomization        : in Integer := 0;
     Preserve             : in Integer := 1) is
    Inf1 : Integer;
    Sup1 : Integer;
    Inf2 : Integer;
    Sup2 : Integer;
    Quality_Mean, Quality_Std : Real := 0.0;    
  begin
    -- Test de validit et dtermination des bornes
    Pre_Alignment(Sequence1,Sequence2,
		  From1,To1,From2,To2,
		  Inf1,Sup1,Inf2,Sup2);
    -- Traiter les sequences en fonction du modele
    case Sequence1.Model is
      when DNA  =>
	declare
	  Sigma : Nucleotide_Match_Matrix;
	  -- La matrice donnant les scores d'appariements 
	  -- entre deux nuclotides
	begin
	  if Matrix = "" then
	    Sigma := 
	      Standard_Nucleotide_Match_Matrix(Match_Counter(Match_Weight),
					     Match_Counter(Mismatch_Weight));
	  else
	    Sigma := Get_Nucleotide_Match_Matrix(Matrix);
	  end if;
	  --Print_Nucleotide_Match_Matrix(Sigma);
	  Nuc_Global_Alignment(Sequence1.Nucleotides(Inf1..Sup1),
			       Sequence2.Nucleotides(Inf2..Sup2),
			       Alignments,
			       Sigma,
			       Match_Weight,Mismatch_Weight,
			       Gap_Weight,Gap_Length_Weight,
			       Progression,Building,Optimize,
			       With_End_Weight,
			       Similarity_Threshold);
	  -- Raliser la randomization si elle est demande
	  if Randomization > 0 then
	    declare
	      Random_Alignments : Alignments_Access;
	      Random_Sequence : Nucleotidic_Array(Inf2..Sup2) :=
		Sequence2.Nucleotides(Inf2..Sup2);
	      Sum_Quality : Integer := 0;
	      Sum2_Quality : Integer := 0;
	    begin
	      for I in 1..Randomization loop
		-- tirer au sort la seconde squence
		Shuffle(Random_Sequence,Preserve);
		-- raliser l'alignement
		Nuc_Global_Alignment(Sequence1.Nucleotides(Inf1..Sup1),
			       Random_Sequence,
			       Random_Alignments,
			       Sigma,
			       Match_Weight,Mismatch_Weight,
			       Gap_Weight,Gap_Length_Weight,
			       Progression,One,Optimize,
			       With_End_Weight,
			       Similarity_Threshold);
		-- calculer les paramtres
		Sum_Quality := Sum_Quality + 
		  Random_Alignments(Random_Alignments'First).Quality; 
	      	Sum2_Quality := Sum2_Quality +
		  Random_Alignments(Random_Alignments'First).Quality * 
		  Random_Alignments(Random_Alignments'First).Quality; 
	      end loop;
	      Quality_Mean := Real(Sum_Quality) / Real(Randomization);
	      Quality_Std := Sqrt((Real(Sum2_Quality) -
				   Real(Sum_quality * Sum_Quality) / 
				   Real(Randomization)) /
				  Real(Randomization));
	    end;
	  end if;
	end;
      when Protein  =>
	declare
	  Sigma : Protein_Match_Matrix;
	  -- La matrice donnant les scores d'appariements 
	  -- entre deux acides amins 
	begin 
	  if Matrix = "" then
	    Sigma := 
	      Standard_Protein_Match_Matrix(Match_Counter(Match_Weight),
					    Match_Counter(Mismatch_Weight));
	  else
	    Sigma := Get_Protein_Match_Matrix(Matrix);
	  end if;
	  --Print_Protein_Match_Matrix(Sigma);
	  Prot_Global_Alignment(Sequence1.Amino_Acids(Inf1..Sup1),
				Sequence2.Amino_Acids(Inf2..Sup2),
				Alignments,
				Sigma,
				Match_Weight,Mismatch_Weight,
				Gap_Weight,Gap_Length_Weight,
				Progression,Building,Optimize,
				With_End_Weight,
				Similarity_Threshold);
	  -- Raliser la randomization si elle est demande
	  if Randomization > 0 then
	    declare
	      Random_Alignments : Alignments_Access;
	      Random_Sequence : Amino_Acids_Array(Inf2..Sup2) :=
		Sequence2.Amino_Acids(Inf2..Sup2);
	      Sum_Quality : Integer := 0;
	      Sum2_Quality : Integer := 0;
	    begin
	      for I in 1..Randomization loop
		-- tirer au sort la seconde squence
		Shuffle(Random_Sequence);
		-- raliser l'alignement
		Prot_Global_Alignment(Sequence1.Amino_Acids(Inf1..Sup1),
				      Random_Sequence,
				      Random_Alignments,
				      Sigma,
				      Match_Weight,Mismatch_Weight,
				      Gap_Weight,Gap_Length_Weight,
				      Progression,One,Optimize,
				      With_End_Weight,
				      Similarity_Threshold);
		-- calculer les paramtres
		Sum_Quality := Sum_Quality + 
		  Random_Alignments(Random_Alignments'First).Quality; 
	      	Sum2_Quality := Sum2_Quality +
		  Random_Alignments(Random_Alignments'First).Quality * 
		  Random_Alignments(Random_Alignments'First).Quality; 
	      end loop;
	      Quality_Mean := Real(Sum_Quality) / Real(Randomization);
	      Quality_Std := Sqrt((Real(Sum2_Quality) -
		              Real(Sum_quality * Sum_Quality) / 
			      Real(Randomization)) /
		Real(Randomization));
	    end;
	  end if;
	end;
      when Others =>
        raise NOT_IMPLEMENTED;
    end case;
    -- Affecter les scores de moyenne et cart-type des alignements
    for I in Alignments'Range loop
      Alignments(I).Quality_Mean := Quality_Mean;
      Alignments(I).Quality_Std  := Quality_Std;
      if Quality_Std = 0.0 then
        Alignments(I).Quality_Zscore := 0.0;
        Alignments(I).Quality_Sng := 0.0;
      else
        Alignments(I).Quality_Zscore :=
          abs(Real(Alignments(I).Quality) - Quality_Mean) /  Quality_Std;
        Alignments(I).Quality_Sng := 
	  Statistics.Normale(Alignments(I).Quality_Zscore);
      end if;
    end loop; 
    -- Complter les informations pour chaque alignement
    Post_Alignment(Sequence1,Sequence2,Alignments,Matrix);
  end Perform_Global_Alignment;

end Sequence_Alignment.Global;
