--------------------------------------------------------------------------
-- Sequence_Alignment.Local                 Copyrigth ENS ULM 1996-1999 --
-- Pierre VINCENS                   V 00.000                 28/12/1998 --
-- Pierre VINCENS                   V 00.001                 03/03/1999 --
--------------------------------------------------------------------------
with Ada.Unchecked_Deallocation;
with Ada.Exceptions; use Ada.Exceptions;

with Strlib; use  Strlib;
with Fifo;

with Ada.Text_Io; use Ada.Text_Io;

with Type_Specifications; use Type_Specifications;
with Type_Specifications.IO; use Type_Specifications.IO;
use  Type_Specifications.IO.Int_IO;

with Sequence_Specifications; use Sequence_Specifications;
with Sequence_Specifications.Basic; use Sequence_Specifications.Basic;

package body Sequence_Alignment.Utils is

  Max_Line_Length     : constant Integer := 2048;

  procedure Path_Clean(Path : in out Path_Record) is
    -- Rcupration de la mmoire associe  un chemin
  begin
    Free(Path.Path);
  end Path_Clean;

  function Cmp_Path(A,B : in Path_Record) return Boolean is
    -- Comparaison de deux chemins
  begin
    if A.Path'First = B.Path'First and A.Path'Last = B.Path'Last then
      for I in A.Path'Range loop
	if A.Path(I) /= B.Path(I) then
	  return False;
	end if;
      end loop;
      return True;
    else
      return False;
    end if;
  end Cmp_Path;

--------------------------------------------------------------------------
-- Gestion du backtracking
--------------------------------------------------------------------------
  procedure Set_Progression(Progression : in Progression_Method;
			    Progress    : out Progression_Array) is
  -- Dfinition de la progression pour la remonte
  begin
    case Progression is
      when DVH =>
	Progress(1) := Align; 
	Progress(2) := Insert_2;
	Progress(3) := Insert_1;
      when DHV =>
	Progress(1) := Align; 
	Progress(2) := Insert_1;
	Progress(3) := Insert_2;
      when VDH =>
	Progress(1) := Insert_2; 
	Progress(2) := Align;
	Progress(3) := Insert_1;
      when VHD =>
	Progress(1) := Insert_2; 
	Progress(2) := Insert_1;
	Progress(3) := Align;
      when HDV =>
	Progress(1) := Insert_1; 
	Progress(2) := Align;
	Progress(3) := Insert_2;
      when HVD =>
	Progress(1) := Insert_1; 
	Progress(2) := Insert_2;
	Progress(3) := Align;
    end case;
  end Set_Progression;
  
  function Best_Path(Path : in Node; 
		     Best : in Integer;
		     Method : in Optimize_Method)
		     return Response is
  begin
    case Method is
      when No =>
	return Equal;
      when Short =>
	if Path.Length < Best then
	  return Good;
	elsif Path.Length = Best then
	  return Equal;
	else
	  return Bad;
	end if;
      when Large =>
	if Path.Length > Best then
	  return Good;
	elsif Path.Length = Best then
	  return Equal;
	else
	  return Bad;
	end if;
    end case;
  end Best_Path;

--------------------------------------------------------------------------
-- Obtention des matrices des valeurs d'appariements
--------------------------------------------------------------------------
  -- Matrice par dfaut pour les nuclotides
  function Standard_Nucleotide_Match_Matrix(Match_Cost : in Match_Counter;
					    Mismatch_Cost : in Match_Counter)
					    return Nucleotide_Match_Matrix is
    Matrix : Nucleotide_Match_Matrix := (others => (others =>Mismatch_Cost));
  begin
    for I in Nucleotide_Type loop
      -- Tous les nuclotides s'apparient avec eux-mmes
      Matrix(I,I) := Match_Cost;
      -- N s'apparie avec tous le monde
      Matrix('N',I) := Match_Cost;
      Matrix(I,'N') := Match_Cost;
    end loop;
    return Matrix;
  end Standard_Nucleotide_Match_Matrix;

  -- Matrice par dfaut pour les acides amines
  function Standard_Protein_Match_Matrix(Match_Cost : in Match_Counter;
					 Mismatch_Cost : in Match_Counter)
					 return Protein_Match_Matrix is
    -- matrice Blosum62
    Matrix : Protein_Match_Matrix := 
      ('A' => ( 'A' =>   4, 'B' =>  -2, 'C' =>   0, 'D' =>  -2, 
		'E' =>  -1, 'F' =>  -2, 'G' =>   0, 'H' =>  -2, 
		'I' =>  -1, 'K' =>  -1, 'L' =>  -1, 'M' =>  -1, 
		'N' =>  -2, 'P' =>  -1, 'Q' =>  -1, 'R' =>  -1, 
		'S' =>   1, 'T' =>   0, 'V' =>   0, 'W' =>  -3, 
		'X' =>   0, 'Y' =>  -2, 'Z' =>  -1, '*' =>  -4 ), 
       'B' => ( 'A' =>  -2, 'B' =>   4, 'C' =>  -3, 'D' =>   4, 
		'E' =>   1, 'F' =>  -3, 'G' =>  -1, 'H' =>   0, 
		'I' =>  -3, 'K' =>   0, 'L' =>  -4, 'M' =>  -3, 
		'N' =>   3, 'P' =>  -2, 'Q' =>   0, 'R' =>  -1, 
		'S' =>   0, 'T' =>  -1, 'V' =>  -3, 'W' =>  -4, 
		'X' =>  -1, 'Y' =>  -3, 'Z' =>   1, '*' =>  -4 ), 
       'C' => ( 'A' =>   0, 'B' =>  -3, 'C' =>   9, 'D' =>  -3, 
		'E' =>  -4, 'F' =>  -2, 'G' =>  -3, 'H' =>  -3, 
		'I' =>  -1, 'K' =>  -3, 'L' =>  -1, 'M' =>  -1, 
		'N' =>  -3, 'P' =>  -3, 'Q' =>  -3, 'R' =>  -3, 
		'S' =>  -1, 'T' =>  -1, 'V' =>  -1, 'W' =>  -2, 
		'X' =>  -2, 'Y' =>  -2, 'Z' =>  -3, '*' =>  -4 ), 
       'D' => ( 'A' =>  -2, 'B' =>   4, 'C' =>  -3, 'D' =>   6, 
		'E' =>   2, 'F' =>  -3, 'G' =>  -1, 'H' =>  -1, 
		'I' =>  -3, 'K' =>  -1, 'L' =>  -4, 'M' =>  -3, 
		'N' =>   1, 'P' =>  -1, 'Q' =>   0, 'R' =>  -2, 
		'S' =>   0, 'T' =>  -1, 'V' =>  -3, 'W' =>  -4, 
		'X' =>  -1, 'Y' =>  -3, 'Z' =>   1, '*' =>  -4 ), 
       'E' => ( 'A' =>  -1, 'B' =>   1, 'C' =>  -4, 'D' =>   2, 
		'E' =>   5, 'F' =>  -3, 'G' =>  -2, 'H' =>   0, 
		'I' =>  -3, 'K' =>   1, 'L' =>  -3, 'M' =>  -2, 
		'N' =>   0, 'P' =>  -1, 'Q' =>   2, 'R' =>   0, 
		'S' =>   0, 'T' =>  -1, 'V' =>  -2, 'W' =>  -3, 
		'X' =>  -1, 'Y' =>  -2, 'Z' =>   4, '*' =>  -4 ), 
       'F' => ( 'A' =>  -2, 'B' =>  -3, 'C' =>  -2, 'D' =>  -3, 
		'E' =>  -3, 'F' =>   6, 'G' =>  -3, 'H' =>  -1, 
		'I' =>   0, 'K' =>  -3, 'L' =>   0, 'M' =>   0, 
		'N' =>  -3, 'P' =>  -4, 'Q' =>  -3, 'R' =>  -3, 
		'S' =>  -2, 'T' =>  -2, 'V' =>  -1, 'W' =>   1, 
		'X' =>  -1, 'Y' =>   3, 'Z' =>  -3, '*' =>  -4 ), 
       'G' => ( 'A' =>   0, 'B' =>  -1, 'C' =>  -3, 'D' =>  -1, 
		'E' =>  -2, 'F' =>  -3, 'G' =>   6, 'H' =>  -2, 
		'I' =>  -4, 'K' =>  -2, 'L' =>  -4, 'M' =>  -3, 
		'N' =>   0, 'P' =>  -2, 'Q' =>  -2, 'R' =>  -2, 
		'S' =>   0, 'T' =>  -2, 'V' =>  -3, 'W' =>  -2, 
		'X' =>  -1, 'Y' =>  -3, 'Z' =>  -2, '*' =>  -4 ), 
       'H' => ( 'A' =>  -2, 'B' =>   0, 'C' =>  -3, 'D' =>  -1, 
		'E' =>   0, 'F' =>  -1, 'G' =>  -2, 'H' =>   8, 
		'I' =>  -3, 'K' =>  -1, 'L' =>  -3, 'M' =>  -2, 
		'N' =>   1, 'P' =>  -2, 'Q' =>   0, 'R' =>   0, 
		'S' =>  -1, 'T' =>  -2, 'V' =>  -3, 'W' =>  -2, 
		'X' =>  -1, 'Y' =>   2, 'Z' =>   0, '*' =>  -4 ), 
       'I' => ( 'A' =>  -1, 'B' =>  -3, 'C' =>  -1, 'D' =>  -3, 
		'E' =>  -3, 'F' =>   0, 'G' =>  -4, 'H' =>  -3, 
		'I' =>   4, 'K' =>  -3, 'L' =>   2, 'M' =>   1, 
		'N' =>  -3, 'P' =>  -3, 'Q' =>  -3, 'R' =>  -3, 
		'S' =>  -2, 'T' =>  -1, 'V' =>   3, 'W' =>  -3, 
		'X' =>  -1, 'Y' =>  -1, 'Z' =>  -3, '*' =>  -4 ),
       'K' => ( 'A' =>  -1, 'B' =>   0, 'C' =>  -3, 'D' =>  -1, 
       		'E' =>   1, 'F' =>  -3, 'G' =>  -2, 'H' =>  -1, 
       		'I' =>  -3, 'K' =>   5, 'L' =>  -2, 'M' =>  -1, 
		'N' =>   0, 'P' =>  -1, 'Q' =>   1, 'R' =>   2,
		'S' =>   0, 'T' =>  -1, 'V' =>  -2, 'W' =>  -3, 
		'X' =>  -1, 'Y' =>  -2, 'Z' =>   1, '*' =>  -4 ), 
       'L' => ( 'A' =>  -1, 'B' =>  -4, 'C' =>  -1, 'D' =>  -4, 
		'E' =>  -3, 'F' =>   0, 'G' =>  -4, 'H' =>  -3, 
		'I' =>   2, 'K' =>  -2, 'L' =>   4, 'M' =>   2, 
		'N' =>  -3, 'P' =>  -3, 'Q' =>  -2, 'R' =>  -2, 
		'S' =>  -2, 'T' =>  -1, 'V' =>   1, 'W' =>  -2, 
		'X' =>  -1, 'Y' =>  -1, 'Z' =>  -3, '*' =>  -4 ), 
       'M' => ( 'A' =>  -1, 'B' =>  -3, 'C' =>  -1, 'D' =>  -3, 
		'E' =>  -2, 'F' =>   0, 'G' =>  -3, 'H' =>  -2, 
		'I' =>   1, 'K' =>  -1, 'L' =>   2, 'M' =>   5, 
		'N' =>  -2, 'P' =>  -2, 'Q' =>   0, 'R' =>  -1, 
		'S' =>  -1, 'T' =>  -1, 'V' =>   1, 'W' =>  -1, 
		'X' =>  -1, 'Y' =>  -1, 'Z' =>  -1, '*' =>  -4 ), 
       'N' => ( 'A' =>  -2, 'B' =>   3, 'C' =>  -3, 'D' =>   1,
		'E' =>   0, 'F' =>  -3, 'G' =>   0, 'H' =>   1,
		'I' =>  -3, 'K' =>   0, 'L' =>  -3, 'M' =>  -2,
		'N' =>   6, 'P' =>  -2, 'Q' =>   0, 'R' =>   0, 
		'S' =>   1, 'T' =>   0, 'V' =>  -3, 'W' =>  -4, 
		'X' =>  -1, 'Y' =>  -2, 'Z' =>   0, '*' =>  -4 ), 
       'P' => ( 'A' =>  -1, 'B' =>  -2, 'C' =>  -3, 'D' =>  -1, 
		'E' =>  -1, 'F' =>  -4, 'G' =>  -2, 'H' =>  -2, 
		'I' =>  -3, 'K' =>  -1, 'L' =>  -3, 'M' =>  -2, 
		'N' =>  -2, 'P' =>   7, 'Q' =>  -1, 'R' =>  -2, 
		'S' =>  -1, 'T' =>  -1, 'V' =>  -2, 'W' =>  -4, 
		'X' =>  -2, 'Y' =>  -3, 'Z' =>  -1, '*' =>  -4 ), 
       'Q' => ( 'A' =>  -1, 'B' =>   0, 'C' =>  -3, 'D' =>   0, 
		'E' =>   2, 'F' =>  -3, 'G' =>  -2, 'H' =>   0, 
		'I' =>  -3, 'K' =>   1, 'L' =>  -2, 'M' =>   0, 
		'N' =>   0, 'P' =>  -1, 'Q' =>   5, 'R' =>   1, 
		'S' =>   0, 'T' =>  -1, 'V' =>  -2, 'W' =>  -2, 
		'X' =>  -1, 'Y' =>  -1, 'Z' =>   3, '*' =>  -4 ), 
       'R' => ( 'A' =>  -1, 'B' =>  -1, 'C' =>  -3, 'D' =>  -2,
		'E' =>   0, 'F' =>  -3, 'G' =>  -2, 'H' =>   0, 
		'I' =>  -3, 'K' =>   2, 'L' =>  -2, 'M' =>  -1, 
		'N' =>   0, 'P' =>  -2, 'Q' =>   1, 'R' =>   5, 
		'S' =>  -1, 'T' =>  -1, 'V' =>  -3, 'W' =>  -3, 
		'X' =>  -1, 'Y' =>  -2, 'Z' =>   0, '*' =>  -4 ), 
       'S' => ( 'A' =>   1, 'B' =>   0, 'C' =>  -1, 'D' =>   0, 
		'E' =>   0, 'F' =>  -2, 'G' =>   0, 'H' =>  -1, 
		'I' =>  -2, 'K' =>   0, 'L' =>  -2, 'M' =>  -1, 
		'N' =>   1, 'P' =>  -1, 'Q' =>   0, 'R' =>  -1, 
		'S' =>   4, 'T' =>   1, 'V' =>  -2, 'W' =>  -3, 
		'X' =>   0, 'Y' =>  -2, 'Z' =>   0, '*' =>  -4 ), 
       'T' => ( 'A' =>   0, 'B' =>  -1, 'C' =>  -1, 'D' =>  -1, 
		'E' =>  -1, 'F' =>  -2, 'G' =>  -2, 'H' =>  -2, 
		'I' =>  -1, 'K' =>  -1, 'L' =>  -1, 'M' =>  -1, 
		'N' =>   0, 'P' =>  -1, 'Q' =>  -1, 'R' =>  -1, 
		'S' =>   1, 'T' =>   5, 'V' =>   0, 'W' =>  -2, 
		'X' =>   0, 'Y' =>  -2, 'Z' =>  -1, '*' =>  -4 ), 
       'V' => ( 'A' =>   0, 'B' =>  -3, 'C' =>  -1, 'D' =>  -3, 
		'E' =>  -2, 'F' =>  -1, 'G' =>  -3, 'H' =>  -3, 
		'I' =>   3, 'K' =>  -2, 'L' =>   1, 'M' =>   1, 
		'N' =>  -3, 'P' =>  -2, 'Q' =>  -2, 'R' =>  -3, 
		'S' =>  -2, 'T' =>   0, 'V' =>   4, 'W' =>  -3, 
		'X' =>  -1, 'Y' =>  -1, 'Z' =>  -2, '*' =>  -4 ), 
       'W' => ( 'A' =>  -3, 'B' =>  -4, 'C' =>  -2, 'D' =>  -4, 
		'E' =>  -3, 'F' =>   1, 'G' =>  -2, 'H' =>  -2, 
		'I' =>  -3, 'K' =>  -3, 'L' =>  -2, 'M' =>  -1, 
		'N' =>  -4, 'P' =>  -4, 'Q' =>  -2, 'R' =>  -3, 
		'S' =>  -3, 'T' =>  -2, 'V' =>  -3, 'W' =>  11,
		'X' =>  -2, 'Y' =>   2, 'Z' =>  -3, '*' =>  -4 ), 
       'X' => ( 'A' =>   0, 'B' =>  -1, 'C' =>  -2, 'D' =>  -1, 
		'E' =>  -1, 'F' =>  -1, 'G' =>  -1, 'H' =>  -1, 
		'I' =>  -1, 'K' =>  -1, 'L' =>  -1, 'M' =>  -1, 
		'N' =>  -1, 'P' =>  -2, 'Q' =>  -1, 'R' =>  -1, 
		'S' =>   0, 'T' =>   0, 'V' =>  -1, 'W' =>  -2, 
		'X' =>  -1, 'Y' =>  -1, 'Z' =>  -1, '*' =>  -4 ), 
       'Y' => ( 'A' =>  -2, 'B' =>  -3, 'C' =>  -2, 'D' =>  -3, 
		'E' =>  -2, 'F' =>   3, 'G' =>  -3, 'H' =>   2, 
		'I' =>  -1, 'K' =>  -2, 'L' =>  -1, 'M' =>  -1, 
		'N' =>  -2, 'P' =>  -3, 'Q' =>  -1, 'R' =>  -2, 
		'S' =>  -2, 'T' =>  -2, 'V' =>  -1, 'W' =>   2, 
		'X' =>  -1, 'Y' =>   7, 'Z' =>  -2, '*' =>  -4 ), 
       'Z' => ( 'A' =>  -1, 'B' =>   1, 'C' =>  -3, 'D' =>   1, 
		'E' =>   4, 'F' =>  -3, 'G' =>  -2, 'H' =>   0, 
		'I' =>  -3, 'K' =>   1, 'L' =>  -3, 'M' =>  -1, 
		'N' =>   0, 'P' =>  -1, 'Q' =>   3, 'R' =>   0, 
		'S' =>   0, 'T' =>  -1, 'V' =>  -2, 'W' =>  -3, 
		'X' =>  -1, 'Y' =>  -2, 'Z' =>   4, '*' =>  -4 ), 
       '*' => ( 'A' =>  -4, 'B' =>  -4, 'C' =>  -4, 'D' =>  -4, 
		'E' =>  -4, 'F' =>  -4, 'G' =>  -4, 'H' =>  -4, 
		'I' =>  -4, 'K' =>  -4, 'L' =>  -4, 'M' =>  -4, 
		'N' =>  -4, 'P' =>  -4, 'Q' =>  -4, 'R' =>  -4, 
		'S' =>  -4, 'T' =>  -4, 'V' =>  -4, 'W' =>  -4, 
		'X' =>  -4, 'Y' =>  -4, 'Z' =>  -4, '*' =>   1 )); 
  begin
    return Matrix;
  end Standard_Protein_Match_Matrix;

  function Get_Match_Matrix(Name : in String) return Match_Matrix is
    Invalid   : constant Match_Counter := Match_Counter'First;
    Matrix    : Match_Matrix := (others => (others => Invalid));
    Min       : Match_Counter := Match_Counter'Last;
    Input     : Ada.Text_Io.File_Type;
    Line      : String(1..Max_Line_Length);
    Length    : Natural;
    type Ordonnee_Array is array(1..30) of Index;
    Ordonnee_Length : Natural := 0;
    Ordonnee : Ordonnee_Array;
    Position : Integer;
    Indice : Index;
  begin
    -- Format des donnes
    -- On ignore les lignes commencant par #
    -- On attend en premier une ligne donnant l'ordre des nuclotides
    -- dans la matrices, puis pour chaque acide amin ses valeurs
    -- soit:
    -- # Blabla blabla
    --     A  B  C ...
    --  A aa ab ac ...
    --  B ba bb bc ...
    --  ...
    --
    -- Ouverture du fichier
    begin
      Open(Input,In_File,Name);
    exception
      when Name_Error =>
	Raise_Exception(Matrix_Name_Error'Identity,Name);
    end;
    -- Lecture des donnes 
    while not End_Of_File(Input) loop
      Get_Line(Input,Line,Length);
      if Length > 0 and then not Is_In_List(Line(1),"#") then
	Position := 1;
	if Ordonnee_Length = 0 then
	  -- Il faut lire la ligne de reference
	  loop
	    while Position <= Length and Line(Position) = ' ' loop
	      Position := Position + 1;
	    end loop;
	    exit when Position > Length;
	    -- Lecture de l'acide amin ou du nuclotide
	    Ordonnee_Length := Ordonnee_Length + 1;
	    Ordonnee(Ordonnee_Length) := To_Index(Line(Position));
	    Position := Position + 1;
	  end loop;
	else
	  -- La ligne de rfrence a t lue
	  -- Lecture de l'indice
	  while Position <= Length and Line(Position) = ' ' loop
	    Position := Position + 1;
	  end loop;
	  if Position <= Length then
	    -- Lecture de l'acide amin, sinon ligne vide
	    Indice := To_Index(Line(Position));
	    -- lecture des valeurs
	    for I in 1..Ordonnee_Length loop
	      Get(Line(Position+1..Length),
		  Matrix(Indice,Ordonnee(I)),
		  Position);
	      Min := Match_Counter'Min(Min,Matrix(Indice,Ordonnee(I)));
	    end loop;
	  end if;
	end if;
      end if;
    end loop;
    -- Toutes les cases de la matrice qui n'ont pas t remplies
    for I in Matrix'Range(1) loop
      for J in Matrix'Range(2) loop
	if Matrix(I,J) = Invalid then
	  -- On regarde si la symtrique est remplie
	  if Matrix(J,I) = Invalid then
	    -- Ce n'est pas le cas, on affecte la valeur minimale
	    Matrix(I,J) := Min;
	  else
	    -- C'est le cas, on reprend cette valeur
	    Matrix(I,J) := Matrix(J,I);
	  end if;
	end if;
      end loop;
    end loop;
    -- Fermeture du fichier
    Close(Input);
    --
    return Matrix;
  end Get_Match_Matrix;

--   procedure Print_Match_Matrix(Sigma : in Match_Matrix) is
--   begin
--     -- Pour debug affichage de la matrice
--     for I in Match_Matrix'Range(1) loop
--       Put(Index'Image(I));
--       Put(" => ( ");
--       for J in Match_Matrix'Range(2) loop
-- 	Put(Index'Image(J));
-- 	Put(" => ");
-- 	Put(Sigma(I,J),3);
-- 	Put(", ");
--       end loop;
--       Put("), ");
--       New_Line;
--     end loop;
--   end Print_Match_Matrix;

  procedure Print_Match_Matrix(Sigma : in Match_Matrix) is
  begin
    -- Pour debug affichage de la matrice
    Put("  ");
    for J in Match_Matrix'Range(2) loop
      Put("  " & To_Character(J));
    end loop;
    New_Line;
    for I in Match_Matrix'Range(1) loop
      Put(To_Character(I));
      Put(" ");
      for J in Match_Matrix'Range(2) loop
	Put(Sigma(I,J),3);
      end loop;
      New_Line;
    end loop;
  end Print_Match_Matrix;

end Sequence_Alignment.Utils;
