--------------------------------------------------------------------------
-- Sequence_Alignment.Utils                 Copyrigth ENS ULM 1996-1999 --
-- Pierre VINCENS                   V 00.000                 28/12/1998 --
-- Pierre VINCENS                   V 00.001                 03/03/1999 --
--------------------------------------------------------------------------
-- Fonctions utilitaires pour l'lignement de squences
--------------------------------------------------------------------------
with Ada.Unchecked_Deallocation;

with Strlib; use  Strlib;
with Fifo;

with Ada.Text_Io; use Ada.Text_Io;
with Type_Specifications.IO; use Type_Specifications.IO;
use  Type_Specifications.IO.Int_IO;

with Sequence_Specifications; use Sequence_Specifications;
with Sequence_Specifications.Basic; use Sequence_Specifications.Basic;

with Sequence_Alignment.Basic; use Sequence_Alignment.Basic;

pragma Elaborate_All(Fifo);

package Sequence_Alignment.Utils is

  
--------------------------------------------------------------------------
-- Structure de donnes associes au calcul des scores et des directions
--------------------------------------------------------------------------
  subtype Progression_Counter is Byte;

  type Direction_Model is (Stop, Align, Insert_1, Insert_2);
    -- Specification de la progression du chemin
    
  type Possible_Direction is array(Direction_Model) of Boolean;
  pragma Pack(Possible_Direction);
    
  type Position_Type is record
    X,Y : Integer;
  end record;
  package Positions_Fifo is new Fifo(Position_Type);
  use Positions_Fifo;
  type Maximal_Score_Type is record
    Stack : Positions_Fifo.Stack_Handler;
    Value : Match_Counter := Match_Counter'First ;
  end record;    
    -- Mmorisation de la valeur du score maximal trouv et des positions
    -- o elles sont localises sous la forme d'une pile FIFO

--------------------------------------------------------------------------
-- Structure de donnes pour la recherche du chemin optimal
--------------------------------------------------------------------------
  type Node is record
    Length      : Integer;   -- Longueur de la squence
    Cross       : Boolean;   -- le point courant ouvre plusieurs choix
    Forward     : Direction_Model; 
                             -- direction choisie pour atteindre le point
                             -- suivant en partant du point courant
    X,Y         : Integer;   -- position atteinte dans le tableau
  end record;

  type Path_Array is array(Integer range <>) of Node;
  -- La structure pour enregistrer un chemin
      
  type Path_Access is access Path_Array;      
  procedure Free is new Ada.Unchecked_Deallocation(Path_Array,Path_Access);
  type Path_Record is record
    Path : Path_Access;
    Progression : Progression_Method;
  end record;
  package Optimal_Path_Fifo is new Fifo(Path_Record);
  use Optimal_Path_Fifo;
  -- On gre les chemins possibles sous la forme d'une pile
  procedure Path_Clean(Path : in out Path_Record);
  procedure Purge is new Optimal_Path_Fifo.Clean_And_Reset(Path_Clean);
  -- La fonction qui permet de dtruire de la mmoire tous les chemins
  -- enregistrs
  function Cmp_Path(A,B : in Path_Record) return Boolean;
  function Is_In_Path_Stack is new Optimal_Path_Fifo.Is_In(Cmp_Path);
  -- Dterminer si un chemin est dj mmoris dans la pile
  type Optimal_Path_Type is record
    Stack : Optimal_Path_Fifo.Stack_Handler;
    Best  : Integer := Integer'First; -- meilleur score trouv
    -- Note: le champ best est introduit afin de pouvoir optimiser
    -- la recherche sur des crit<E8>res diff<E9>rents
  end record;

--------------------------------------------------------------------------
-- Gestion du backtracking
--------------------------------------------------------------------------
  type Progression_Array is array(1..3) of Direction_Model;
    
  procedure Set_Progression(Progression : in  Progression_Method;
			    Progress    : out Progression_Array);
  pragma Inline(Set_Progression);
  -- Construit un tableau donnant l'ordre des progressions

  type Response is (Good,Equal,Bad);
  function Best_Path(Path : in Node; 
		     Best : in Integer;
		     Method : in Optimize_Method)
		     return Response;
  pragma Inline(Best_Path);
  -- Indique si la nouvelle solution est meilleure ou quivalente
  -- en fonction du critre d'optimisation choisi

--------------------------------------------------------------------------
-- Calcul de la matrice des valeurs d'appariements
--------------------------------------------------------------------------
  function Standard_Nucleotide_Match_Matrix(Match_Cost : Match_Counter;
					    Mismatch_Cost : Match_Counter)
					    return Nucleotide_Match_Matrix;
  pragma Inline(Standard_Nucleotide_Match_Matrix);
  -- Calcule la matrice d'appariement standard entre les nuclotides

  function Standard_Protein_Match_Matrix(Match_Cost : Match_Counter;
					 Mismatch_Cost : Match_Counter)
					 return Protein_Match_Matrix;
  pragma Inline(Standard_Protein_Match_Matrix);
  -- Calcule la matrice d'appariement standard entre les acides amins

  MATRIX_NAME_ERROR : exception;
  -- lev si le fichier contenant la matrice de scores n'est pas accessible
  MATRIX_FORMAT_ERROR : exception;
  -- lev si le format du fichier contenant la matrice de scores est
  -- incorrect

  generic
    type Index is (<>);
    type Match_Matrix is array(Index,Index) of Match_Counter;
    with function To_Index(X : in Character) return Index;
  function Get_Match_Matrix(Name : in String) return Match_Matrix;
  pragma Inline(Get_Match_Matrix);
  -- Lecture d'une matrice de scores dans un fichier

  generic
    type Index is (<>);
    type Match_Matrix is array(Index,Index) of Match_Counter;
    with function To_Character(X : in Index) return Character;
  procedure Print_Match_Matrix(Sigma : in Match_Matrix);
  pragma Inline(Print_Match_Matrix);
  -- Affiche une matrice de scores d'appariements

end Sequence_Alignment.Utils;
