--------------------------------------------------------------------------
-- Strlib                                   Copyrigth ENS ULM 1986-1997 --
-- Pierre VINCENS                   V 01.000                 06/11/1997 --
-- Pierre VINCENS                   V 01.001                 08/04/1999 --
-- Pierre VINCENS                   V 01.002                 12/01/2000 --
--------------------------------------------------------------------------
--
-- Paquetage de gestion de chaines de caracteres sous une forme dynamique
-- A l'origine ecrit en Ada83, il est maintenu pour compatibilite avec les
-- applications existantes.
--
-- Note 1: les fonctions d'entres-sorties ont t reportes dans un
-- paquetage fils afin de conserver ce paquetage prellaborable
--
-- Note 2: l'accs  la chaine de caractre se fait  travers un
-- pointeur, dont l'usage est gnralement transparent pour l'utilisateur.
-- ATTENTION cependant, les fonctions crant une nouvelle chaine
-- vrifie pralablement que le pointeur est null. Si ce n'est pas
-- le cas, elle dtruise la chaine prcdente. Cela peut poser
-- des problemes lorsqu'il y a eu affectation d'une variable du type
-- varstring dans une autre. La destruction de la chaine conduit
-- a ce que le pointeur recopi dsigne une adresse mmoire invalide.
-- Ce problme pourrait tre rsolu soit:
--   o en definissant le type varstring comme limited
--   o en specifiant ce type comme controlled avec les procdures
--     d'initialisation, de destruction et d'ajustement adequate
-- Il a t choisi dans cette version de ne pas implmenter ces
-- choix pour ne pas alourdir l'usage de ce paquetage. L'implmentation
-- sous la forme d'un type controlled pourra tre envisag ultrieurement
-- lorsque l'implmentation de ces types aura t amliore. 
--
--------------------------------------------------------------------------
-- V 01.002                
-- Ajout de la fonction Raz pour remettre a null une chaine varstring 
-- sans liberer la place associee. Cette fonction est necessaire, par
-- exemple, lorqu'il y a enregistrement dans une structure dynamique
-- de valeur de ce type. 
--------------------------------------------------------------------------
with Ada.Unchecked_Deallocation;

package Strlib is

  pragma Preelaborate(Strlib);

--===========================================================================
-- Listes des erreurs potentielles
--===========================================================================
  INDEX_ERROR : exception;
    -- levee dans substring si les indices demandes sont incompatibles
    -- levee dans next_word et previous_word si la position de debut
    -- de recherche est en dehors de la chaine.

  STRING_NOT_FOUND : exception;
    -- levee dans search si la chaine n'est pas trouvee

  NULL_STRING : exception;
    -- levee si l'on essaye d'acceder a une chaine non initialisee

--===========================================================================
-- Definition des types de base
--===========================================================================
  type String_Access is access String;
  procedure Free is new Ada.Unchecked_Deallocation(String,String_Access);

  type Varstring is record
    Value  : String_Access;
  end record;

  type End_Mode is (Normal, Removing);
    -- Ce type est utilise pour preciser si on conserve la chaine
    -- telle qu'elle ou si l'on supprime les blancs en fin de chaine

--===========================================================================
-- Fonctionnalites de conversion depuis et vers le type string
--===========================================================================
  function To_Varstring(Str : in String) return Varstring;
  function To_String(Varstr : in Varstring) return String;

--===========================================================================
-- Liberation de la place associee a une chaine
--===========================================================================
  procedure Clear(Varstr : in out Varstring);
  
--===========================================================================
-- Remise a zero (null) d'une variable sans liberation de la place associee
--===========================================================================
  procedure Raz(Varstr : in out Varstring);
  
--===========================================================================
-- Affectation d'une chaine
--===========================================================================
  procedure Copy(Src : in String; Dst : in out Varstring);
  procedure Copy(Src : in Varstring; Dst : in out Varstring);
  -- Recopie d'une chaine complete
  procedure Copy(Src    : in String; 
		 Length : in Positive;
		 Dst    : in out Varstring);
  procedure Copy(Src    : in Varstring; 
		 Length : in Positive;
		 Dst    : in out Varstring);
  -- Recopie d'au maximum length caracteres de la chaine src dans dst
 
--===========================================================================
-- L'operateur de concatenation
--===========================================================================
  function "&"(Str1,Str2 : in Varstring) return Varstring;
  function "&"(Str1 : in String; Str2 : in Varstring) return Varstring;
  function "&"(Str1 : in Varstring; Str2 : in String) return Varstring;
  function "&"(Str1 : in Character; Str2 : in Varstring) return Varstring;
  function "&"(Str1 : in Varstring; Str2 : in Character) return Varstring;

--===========================================================================
-- L'extraction d'une chaine de caracteres
--===========================================================================
  function Substring(Varstr     : in Varstring;
                     First,Last : in Positive) return Varstring;

--===========================================================================
-- La recherche d'une chaine de caracteres
--===========================================================================
  function Search(Str   : in String;
                  Value : in String) return Integer;
  function Search(Varstr  : in Varstring;
                  Value   : in String) return Integer;
  function Search(Varstr  : in Varstring;
                  Value   : in Varstring) return Integer;
  function Search(Str     : in String;
                  Value   : in Varstring) return Integer;
  -- Recherche d'une chaine de caracteres value dans une autre chaine
  -- Si la chaine n'est pas trouve, alors l'exception STRING_NOT_FOUND
  -- est levee. La valeur retournee est l'indice de position du debut
  -- de la chaine.

  function RK_Search_Forward(Str     : in String;
                             Pattern : in String) return Integer;
  function RK_Search_Forward(Varstr  : in Varstring;
                             Pattern : in String) return Integer;
  function RK_Search_Forward(Varstr  : in Varstring;
                             Pattern : in Varstring) return Integer;
  function RK_Search_Forward(Str     : in String;
                             Pattern : in Varstring) return Integer;
  -- Recherche d'une chaine de caractere par la methode de Rabin et Karp
  -- la recherche se fait depuis le debut vers la fin
  -- Reference: Sedgewick p 290
  -- Si la chaine n'est pas trouve, alors l'exception STRING_NOT_FOUND
  -- est levee. La valeur retournee est l'indice de position du debut
  -- de la chaine.

  function RK_Search_Backward(Str     : in String;
                              Pattern : in String) return Integer;
  function RK_Search_Backward(Varstr  : in Varstring;
                              Pattern : in String) return Integer;
  function RK_Search_Backward(Varstr  : in Varstring;
                              Pattern : in Varstring) return Integer;
  function RK_Search_Backward(Str     : in String;
                              Pattern : in Varstring) return Integer;
  -- Recherche d'une chaine de caractere par la methode de Rabin et Karp
  -- la recherche se fait depuis la fin de src vers le debut
  -- Reference: Sedgewick p 290
  -- Si la chaine n'est pas trouve, alors l'exception STRING_NOT_FOUND
  -- est levee. La valeur retournee est l'indice de position du debut
  -- de la chaine.

  function Pattern_Is_In_String(Str     : in String;
                                Pattern : in String) return Boolean;
  function Pattern_Is_In_String(Str     : in Varstring;
                                Pattern : in String) return Boolean;
  function Pattern_Is_In_String(Str     : in Varstring;
                                Pattern : in Varstring) return Boolean;
  function Pattern_Is_In_String(Str     : in String;
                                Pattern : in Varstring) return Boolean;
  -- Cette fonction verifie si ule pattern est present dans la chaine
  -- de caracteres passee en argument. Elle fait appel a RK_Search_Backward.

--===========================================================================
-- La verification de la presence d'un caractere dans une chaine
--===========================================================================
  function Is_In(Value : in Character;
                 Str   : in String) return Boolean;

  function Is_In(Value  : in Character;
                 Varstr : in Varstring) return Boolean;

--===========================================================================
-- La verification de la presence d'une chaine de caracteres
--===========================================================================
  function Exist(Str   : in String;
                 Value : in String) return Boolean;
  function Exist(Varstr  : in Varstring;
                 Value   : in String) return Boolean;
  function Exist(Varstr  : in Varstring;
                 Value   : in Varstring) return Boolean;
  function Exist(Str     : in String;
                 Value   : in Varstring) return Boolean;

--===========================================================================
-- La lecture par mot
--===========================================================================
  procedure Next_Word(Str      : in String;
                      Word     : in out Varstring;
                      Position : in out Integer;
                      Pattern  : in String := " :.,?!");
  procedure Next_Word(Str      : in Varstring;
                      Word     : in out Varstring;
                      Position : in out Integer;
                      Pattern  : in String := " :.,?!");
  -- Cette procedure recherche le prochain mot dans la chaine de caractere
  -- str. Position donne la position de dpart de recherche dans la chaine
  -- la valeur retourne est la position de fin du mot incremente de 1.
  -- Pattern est une chaine de caractres ou sont prciss les caractres
  -- sparateur de mots.
  -- Lexception INDEX_ERROR est levee si position est en dehors des
  -- bornes de la chaine.

  procedure Previous_Word(Str      : in String;
                          Word     : in out Varstring;
                          Position : in out Integer;
                          Pattern  : in String := " :.,?!");
  procedure Previous_Word(Str      : in Varstring;
                          Word     : in out Varstring;
                          Position : in out Integer;
                          Pattern  : in String := " :.,?!");
  -- Calque sur la precedente, cette procedure recherche le mot
  -- precedent la position position dans la chaine str.

--===========================================================================
-- L'insertion d'une chaine de caracteres
--===========================================================================
  function Insert(Varstr     : in Varstring;
                  Str        : in String;
                  Position   : in Positive) return Varstring;
  function Insert(Varstr     : in Varstring;
                  Str        : in Varstring;
                  Position   : in Positive) return Varstring;
  function Append(Varstr     : in Varstring;
                  Str        : in String;
                  Position   : in Positive) return Varstring;
  function Append(Varstr     : in Varstring;
                  Str        : in Varstring;
                  Position   : in Positive) return Varstring;


--===========================================================================
-- Deletion d'une portion de chaine de caracteres
--===========================================================================
  function Delete(Varstr : in Varstring;
                  First  : in Positive;
                  Last   : in Positive) return Varstring;

--===========================================================================
-- comparaison de 2 chaines de caracteres
--===========================================================================
  function Equal(Str1, Str2 : in Varstring) return Boolean;
  function Equal(Str1 : in String; Str2 : in Varstring) return Boolean;
  function Equal(Str1 : in Varstring; Str2 : in String) return Boolean;
  function Diff(Str1, Str2 : in Varstring) return Boolean;
  function Diff(Str1 : in String; Str2 : in Varstring) return Boolean;
  function Diff(Str1 : in Varstring; Str2 : in String) return Boolean;

--===========================================================================
-- Renverser une chaine de caracteres
--===========================================================================
  function Invert(Varstr     : in Varstring) return Varstring;
  procedure Invert(Varstr : in out Varstring);

--===========================================================================
-- Ecrire dans une chaine de caractres les N derniers caractres
--===========================================================================
  procedure Put_The_N_Last_Characters(Output : out String;
                                      Str    : in String);
  procedure Put_The_N_Last_Characters(Output : out String;
				    Varstr : in Varstring);
  -- Des fonctions quivalentes existent pour crire dans un fichier
  -- Voir Strlib.Io

--===========================================================================
-- Quelques fonctionnalites de service
--===========================================================================
  function Length(Varstr     : in Varstring) return Natural;

end Strlib;
